__doc__ = """

.. autoclass:: miv.core.pipeline.Pipeline
   :members:

"""
__all__ = ["Pipeline"]

from typing import cast
from collections.abc import Sequence

import os
import pathlib
import time

from miv.core.operator.protocol import _Node
from miv.core.utils.graph_sorting import topological_sort


class Pipeline:
    """
    A pipeline is a collection of operators that are executed in a specific order.

    If operator structure is a tree like

    .. mermaid::

        flowchart LR
            A --> B --> D --> F
            A --> C --> E --> F
            B --> E

    then the execution order of `Pipeline(F)` is A->B->D->C->E->F.
    If nodes already have cached results, they will be loaded instead of being executed.
    For example, if E is already cached, then the execution order of `Pipeline(F)` is A->B->D->F. (C is skipped, E is loaded from cache)
    """

    def __init__(self, node: _Node | Sequence[_Node]) -> None:
        self.nodes_to_run: list[_Node]
        if not isinstance(node, list):
            # FIXME: check if the node is standalone operator
            node = cast(_Node, node)
            self.nodes_to_run = [node]
        else:
            node = cast(Sequence[_Node], node)
            self.nodes_to_run = list(node)

    def run(
        self,
        working_directory: str | pathlib.Path = "./results",
        cache_directory: str | pathlib.Path | None = None,
        temporary_directory: str | pathlib.Path | None = None,
        skip_plot: bool = False,
        verbose: bool = False,  # Use logging
    ) -> None:
        """
        Run the pipeline.

        Parameters
        ----------
        working_directory : Optional[Union[str, pathlib.Path]], optional
            The working directory where the pipeline will be executed. By default "./results"
        cache_directory : Optional[Union[str, pathlib.Path]], optional
            The cache directory where the pipeline will be executed. By default None
            If None, the cache directory will be the same as the working directory.
        temporary_directory : Optional[Union[str, pathlib.Path]], optional
            If given, files will be saved in temporary directory, and will be moved to
            working directory after. Cache directory is not altered.
        verbose : bool, optional
            If True, the pipeline will log debugging informations. By default False
        """
        # Set working directory
        if cache_directory is None:
            cache_directory = working_directory

        # Reset all callbacks
        for last_node in self.nodes_to_run:
            for node in topological_sort(last_node):
                if hasattr(node, "reset_callbacks"):
                    node.reset_callbacks(plot=skip_plot)
                if hasattr(node, "set_save_path"):
                    if temporary_directory is not None:
                        node.set_save_path(temporary_directory, cache_directory)
                    else:
                        node.set_save_path(working_directory, cache_directory)

        # Execute
        for node in self.nodes_to_run:
            if verbose:
                stime = time.time()
                print("Running: ", node, flush=True)

            try:
                node.output()
            except Exception as e:
                print("  Exception raised: ", node, flush=True)
                raise e

            if verbose:
                etime = time.time()
                print(f"  Finished: {etime - stime:.03f} sec")
                print("Pipeline done.", flush=True)

        if temporary_directory is not None:
            os.system(f"cp -rf {temporary_directory}/* {working_directory}/")
            # import shutil
            # shutil.move(temporary_directory, working_directory)

    def summarize(self) -> str:
        strs = []
        for node in self.nodes_to_run:
            execution_order = topological_sort(node)

            strs.append(f"Execution order for {node}:")
            for i, op in enumerate(execution_order):
                strs.append(f"{i}: {op}")
        return "\n".join(strs)



__doc__ = """

Useful wrapper functions for MIV operators.

.. autofunction:: miv.core.wrapper.cache_call

.. autofunction:: miv.core.wrapper.cached_method

.. autofunction:: miv.core.operator_generator.wrapper.cache_generator_call


"""

__all__ = [
    "cache_call",
    "cached_method",
]

from typing import Any, TypeVar, Concatenate, ParamSpec, cast
from collections.abc import Callable
import os
import functools
from joblib import Memory

from .protocol import _Cachable

F = TypeVar("F")
P = ParamSpec("P")
R = TypeVar("R")
SELF = TypeVar("SELF", bound=_Cachable)


def cache_call(func: Callable[..., F]) -> Callable[..., F]:
    """
    Cache the methods of the operator.
    Save the cache in the cacher object.
    """

    def wrapper(self: _Cachable, *args: Any, **kwargs: Any) -> F:
        tag = "data"
        cacher = self.cacher

        result = func(self, *args, **kwargs)
        if result is None:
            # In case the module does not return anything
            return cast(F, None)
        cacher.save_cache(result, tag=tag)
        cacher.save_config(tag=tag)
        return result

    return wrapper


def cached_method(
    cache_tag: str | None = None,
    *,
    verbose: bool = False,
) -> Callable[[Callable[Concatenate[SELF, P], R]], Callable[Concatenate[SELF, P], R]]:
    """
    Cache the results of instance methods.
    Uses joblib's Memory to handle caching with optional custom tag and verbosity.

    Example
    -------
        >>> from miv.core.operator.wrapper import cached_method
        >>> from miv.core.operator.protocol import _Cachable
        >>>
        >>> class MyOperator(OperatorMixin):
        ...     def __init__(self):
        ...         self.cacher: _CacherProtocol
        ...
        ...     @cached_method("my_tag")  # Cache with custom tag. Otherwise, it uses the function name.
        ...     def expensive_computation(self, x: int) -> int:
        ...         # This will only be computed once for each unique input
        ...         return x * x
        ...
        >>> op = MyOperator()
        >>> result = op.expensive_computation(5)  # Computes and caches
        >>> result = op.expensive_computation(5)  # Returns cached result
    """

    def decorator(
        func: Callable[Concatenate[SELF, P], R],
    ) -> Callable[Concatenate[SELF, P], R]:
        @functools.wraps(func)
        def wrapper(self: SELF, *args: P.args, **kwargs: P.kwargs) -> R:
            # Check if caching is allowed based on policy
            policy = self.cacher.policy
            if policy == "OFF":
                # Never use cache, always run the function
                return func(self, *args, **kwargs)

            # For other policies, we'll use the caching mechanism
            cached_methods = getattr(self, "_cached_methods", {})
            cache_dir = self.cacher.cache_dir
            tag = cache_tag if cache_tag is not None else func.__name__

            if tag not in cached_methods:
                cache_path = os.path.join(cache_dir, tag)
                memory = Memory(cache_path, verbose=verbose)
                _func = memory.cache(
                    func, ignore=["self"]
                )  # ignore is used to bypass un-hashable self
                cached_methods[tag] = _func
                setattr(self, "_cached_methods", cached_methods)  # noqa
            if policy == "MUST":
                if not cached_methods[tag].check_call_in_cache(*args, **kwargs):  # type: ignore
                    raise RuntimeError(
                        f"MUST policy is used for caching, but cache for {tag} does not exist."
                    )
            elif policy == "OVERWRITE":
                cached_methods[tag].clear(warn=False)  # type: ignore

            return cached_methods[tag](self, *args, **kwargs)  # type: ignore

        return wrapper  # type: ignore

    return decorator



from typing import Any

from miv.core.datatype.events import *
from .pure_python import *
from .signal import *
from .spikestamps import *

DataTypes = Any  # Union[ # TODO
#    miv.core.datatype.signal.Signal,
#    miv.core.datatype.signal.Spikestamps,
# ]



__doc__ = """
Connectivity module
"""
__all__ = ["DirectedConnectivity", "UndirectedConnectivity"]

import csv
import functools
import itertools
import logging
import os
import pathlib
from dataclasses import dataclass
from typing import Any

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import scipy.stats as spst
from elephant.causality.granger import pairwise_granger
from tqdm import tqdm

from miv.core.datatype import Signal, Spikestamps
from miv.core.operator.operator import OperatorMixin
from miv.core.operator.wrapper import cache_call
from miv.mea import mea_map
from miv.statistics.spiketrain_statistics import firing_rates


# self MPI-able
@dataclass
class DirectedConnectivity(OperatorMixin):
    """
    Directional connectivity analysis operator

    Parameters
    ----------
    bin_size : pq.Quantity
        Bin size for spike train
    mea: Optional[Union[str, np.ndarray]]
        2D array map of MEA channel layout.
    tag : str, optional
        Tag for this operator, by default "directional connectivity analysis"
    progress_bar : bool, optional
        Show progress bar, by default True
    skip_surrogate : bool, optional
        Skip surrogate analysis, by default False
    surrogate_N : int, optional
        Number of surrogate, by default 30
    p_threshold : float, optional
        p-value threshold, by default 0.05
    seed : int, optional
        Random seed. If None, use random seed, by default None
    """

    mea: str = None
    channels: list[int] | None = None
    exclude_channels: list[int] | None = None
    bin_size: float = 0.001
    minimum_count: int = 1
    tag: str = "directional connectivity analysis"
    progress_bar: bool = False

    skip_surrogate: bool = False

    # Surrogate parameters
    surrogate_N: int = 30
    p_threshold: float = 0.05
    H_threshold: float = 1e-2
    seed: int = None
    num_proc: int = 1

    def __post_init__(self):
        super().__init__()
        if isinstance(self.mea, str):
            self.mea_map = mea_map[self.mea]
        else:
            self.mea_map = mea_map["64_intanRHD"]

        if self.exclude_channels is None:  # FIXME: Use dataclass default value
            self.exclude_channels = []

    @cache_call
    def __call__(self, spikestamps: Spikestamps) -> np.ndarray:
        """__call__.

        Parameters
        ----------
        spikestamps : Spikestamps
        """
        binned_spiketrain: Signal = spikestamps.binning(
            bin_size=self.bin_size, minimum_count=self.minimum_count
        )

        # Channel Selection
        if self.channels is None:
            n_nodes = binned_spiketrain.number_of_channels
            channels = tuple(range(n_nodes))
        else:
            n_nodes = len(self.channels)
            channels = tuple(self.channels)
            binned_spiketrain = binned_spiketrain.select(channels)

        # Get adjacency matrix based on transfer entropy
        adj_matrix = np.zeros([n_nodes, n_nodes], dtype=np.bool_)  # source -> target
        connectivity_metric_matrix = np.zeros(
            [n_nodes, n_nodes], dtype=np.float64
        )  # source -> target

        pairs = [
            (i, j)
            for i, j in itertools.product(range(n_nodes), range(n_nodes))
            if i not in self.exclude_channels and j not in self.exclude_channels
        ]
        func = functools.partial(
            self._get_connection_info,
            binned_spiketrain=binned_spiketrain,
            channels=channels,
            mea=self.mea_map,
            skip_surrogate=self.skip_surrogate,
            surrogate_N=self.surrogate_N,
            seed=self.seed,
            H_threshold=self.H_threshold,
        )
        # with mp.Pool(self.num_proc) as pool:
        #    for idx, result in enumerate(
        #        tqdm(
        #            pool.imap(func, pairs),
        #            total=len(pairs),
        #            disable=not self.progress_bar,
        #        )
        #    ):
        #        i, j = pairs[idx]
        #        adj_matrix[i, j] = result[0] < self.p_threshold
        #        connectivity_metric_matrix[i, j] = result[1]
        pbar = tqdm(total=len(pairs), disable=not self.progress_bar)
        for idx, pair in enumerate(pairs):
            pbar.update(1)
            result = func(pair)
            i, j = pair
            adj_matrix[i, j] = result[0] < self.p_threshold
            connectivity_metric_matrix[i, j] = result[1]

        connection_ratio = adj_matrix.sum() / adj_matrix.ravel().shape[0]

        info = dict(
            adjacency_matrix=adj_matrix,
            connectivity_matrix=connectivity_metric_matrix,
            connection_ratio=connection_ratio,
        )

        return info

    @staticmethod
    def _get_connection_info(
        pair,
        binned_spiketrain,
        channels,
        mea,
        skip_surrogate,
        surrogate_N,
        seed,
        H_threshold,
    ):
        """
        Get connection information
        """
        sid, tid = pair
        if sid == tid:
            return 1, 0
        if channels[sid] not in mea or channels[tid] not in mea:
            return 1, 0
        source = binned_spiketrain[sid]
        target = binned_spiketrain[tid]
        te, surrogate_te_list = DirectedConnectivity._surrogate_t_test(
            source,
            target,
            skip_surrogate=skip_surrogate,
            surrogate_N=surrogate_N,
            seed=seed,
        )
        if te < H_threshold:
            return 1, 0
        return 1, te
        t_value, p_value = spst.ttest_1samp(surrogate_te_list, te, nan_policy="omit")
        return p_value, te

    def plot_adjacency_matrix(self, result, inputs, save_path=None, show=False):
        connectivity_metric_matrix = result["connectivity_matrix"]

        fig, ax = plt.subplots(figsize=(12, 12))
        im = ax.imshow(connectivity_metric_matrix, cmap="gray_r", vmin=0)
        ax.set_xlabel("Target")
        ax.set_ylabel("Source")
        ax.set_title("Transfer Entropy")
        plt.colorbar(im, ax=ax)

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "te.png"))

        if show:
            plt.show()

        plt.close(fig)

    def plot_transfer_entropy_histogram(
        self, result, inputs, save_path=None, show=False
    ):
        connectivity_metric_matrix = result["connectivity_matrix"]

        # Export values in csv
        if save_path is not None:
            with open(os.path.join(save_path, "te_values.csv"), mode="w") as f:
                writer = csv.writer(f)
                writer.writerow(["source", "target", "metrics"])

                for i, row in enumerate(connectivity_metric_matrix):
                    for j, value in enumerate(row):
                        writer.writerow([i, j, value])

        # Plot values in heatmap
        fig, ax = plt.subplots(figsize=(12, 12))
        ax.hist(connectivity_metric_matrix, bins=30)
        ax.set_xlabel("transfer entropy")
        ax.set_ylabel("count")
        ax.set_title("Transfer Entropy Histogram")

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "te_histogram.png"))

        if show:
            plt.show()

        plt.close(fig)

    def plot_nodewise_connectivity(
        self,
        result: Any,
        inputs,
        save_path: str | pathlib.Path = None,
        show: bool = False,
    ):
        """
        Plot nodewise connectivity

        Parameters
        ----------
        result : Any
            Result of this operation.
        save_path : str, optional
            Save path, by default None
        show : bool, optional
            Show plot, by default False

        """
        return
        if show:
            logging.warning(
                "show is not supported for node-wise connectivity plot. Plots will be saved only, if save_path is specified."
            )
            show = False
        if save_path is None:
            return
        adj_matrix = result["adjacency_matrix"]
        connectivity_metric_matrix = result["connectivity_matrix"]

        n_nodes = adj_matrix.shape[0]
        channels = self.channels if self.channels is not None else list(range(n_nodes))

        # Save connectivity plot
        for idx, source in enumerate(channels):
            if source not in self.mea_map:
                continue
            self._plot_directionality(
                adj_matrix[idx],
                source,
                os.path.join(save_path, f"p_graph_source_{source}.png"),
            )
            self._plot_directionality(
                connectivity_metric_matrix[idx],
                source,
                os.path.join(save_path, f"te_graph_source_{source}.png"),
                boolean_connectivity=False,
            )

        for idx, target in enumerate(channels):
            if target not in self.mea_map:
                continue
            self._plot_directionality(
                adj_matrix[:, idx],
                target,
                os.path.join(save_path, f"p_graph_target_{target}.png"),
                reverse=True,
            )
            self._plot_directionality(
                connectivity_metric_matrix[:, idx],
                target,
                os.path.join(save_path, f"te_graph_target_{target}.png"),
                reverse=True,
                boolean_connectivity=False,
            )

    def _plot_directionality(
        self,
        connectivity,
        self_index,
        savename,
        reverse=False,
        boolean_connectivity=True,
    ):
        """
        Plot directionality
        """
        G_1 = nx.DiGraph()

        # position
        pos = {}
        for i in range(self.mea_map.shape[0]):
            for j in range(self.mea_map.shape[1]):
                center = self.mea_map[i, j]
                if center < 0:
                    continue
                G_1.add_node(center)
                pos[center] = (j, -i)

        if boolean_connectivity:
            for idx, conn in enumerate(connectivity):
                if idx == self_index:
                    continue  # No self-connection
                if idx not in self.mea_map:
                    continue  # No connection if node does not exist
                if not conn:
                    continue  # No connection

                if not reverse:
                    G_1.add_edge(self_index, idx)
                else:
                    G_1.add_edge(idx, self_index)
        else:
            for idx, conn in enumerate(connectivity):
                if idx == self_index:
                    continue  # No self-connection
                if idx not in self.mea_map:
                    continue  # No connection if node does not exist

                if not reverse:
                    G_1.add_edge(self_index, idx, weight=conn)
                else:
                    G_1.add_edge(idx, self_index, weight=conn)

        # Numbering
        # edge_weights = [G_1[u][v]['weight'] if 'weight' in G_1[u][v] else 0.0 for u, v in G_1.edges()]
        plt.figure()
        widths = nx.get_edge_attributes(G_1, "weight")
        nx.draw(G_1, pos)
        if boolean_connectivity:
            nx.draw_networkx_edges(
                G_1,
                pos,
                edgelist=widths.keys(),
                ax=plt.gca(),
            )
        else:
            nx.draw_networkx_edges(
                G_1,
                pos,
                edgelist=widths.keys(),
                edge_color=list(widths.values()),
                edge_cmap=plt.cm.gray_r,
                edge_vmin=0.0,
                edge_vmax=1.0,
                ax=plt.gca(),
            )
        nx.draw_networkx_labels(
            G_1,
            pos,
            labels=dict(zip(G_1.nodes(), G_1.nodes(), strict=False)),
            font_color="white",
            ax=plt.gca(),
        )
        # for node, (x_coord, y_coord) in pos.items():
        #    plt.text(x_coord - 0.1, y_coord - 0.1, str(node))

        plt.savefig(savename)
        plt.close()


@dataclass
class UndirectedConnectivity(OperatorMixin):
    """
    Undirected connectivity analysis operator
    Runs correlation

    Parameters
    ----------
    bin_size : pq.Quantity
        Bin size for spike train
    tag : str, optional
        Tag for this operator, by default "directional connectivity analysis"
    progress_bar : bool, optional
        Show progress bar, by default True
    skip_surrogate : bool, optional
        Skip surrogate analysis, by default False
    surrogate_N : int, optional
        Number of surrogate, by default 30
    p_threshold : float, optional
        p-value threshold, by default 0.05
    seed : int, optional
        Random seed. If None, use random seed, by default None
    """

    exclude_channels: list[int] | None = None
    bin_size: float = 0.001
    minimum_count: int = 1
    firing_rate_limit: float = 5e-1
    tag: str = "undirected connectivity analysis"
    progress_bar: bool = False

    skip_surrogate: bool = False

    # Surrogate parameters
    surrogate_N: int = 30
    p_threshold: float = 0.05
    seed: int = None
    num_proc: int = 1

    def __post_init__(self):
        super().__init__()
        if self.exclude_channels is None:  # FIXME: Use dataclass default value
            self.exclude_channels = []

    @cache_call
    def __call__(self, spikestamps: Spikestamps, mea=None) -> np.ndarray:
        """__call__.

        Parameters
        ----------
        spikestamps : Spikestamps
        """
        binned_spiketrain: Signal = spikestamps.binning(
            bin_size=self.bin_size, minimum_count=self.minimum_count
        )
        rates = firing_rates(spikestamps)["rates"]

        # Channel Selection
        n_nodes = binned_spiketrain.number_of_channels
        channels = list(range(n_nodes))

        # Get adjacency matrix based on transfer entropy
        adj_matrix = np.zeros([n_nodes, n_nodes], dtype=np.bool_)  # source -> target
        connectivity_metric_matrix = np.zeros(
            [n_nodes, n_nodes], dtype=np.float64
        )  # source -> target

        pairs = [
            (i, j)
            for i, j in itertools.product(range(n_nodes), range(n_nodes))
            if i not in self.exclude_channels
            and j not in self.exclude_channels
            and i < j
            and i != j
            and rates[i] > self.firing_rate_limit
            and rates[j] > self.firing_rate_limit
        ]
        func = functools.partial(
            self._get_connection_info,
            binned_spiketrain=binned_spiketrain,
            channels=channels,
            mea=mea,
            skip_surrogate=self.skip_surrogate,
            surrogate_N=self.surrogate_N,
            seed=self.seed,
        )
        pbar = tqdm(total=len(pairs), disable=not self.progress_bar)
        for idx, pair in enumerate(pairs):
            pbar.update(1)
            result = func(pair)
            i, j = pair
            adj_matrix[i, j] = result[0] < self.p_threshold
            connectivity_metric_matrix[i, j] = result[1][0]
            connectivity_metric_matrix[j, i] = result[1][1]

        connection_ratio = adj_matrix.sum() / adj_matrix.ravel().shape[0]

        info = dict(
            adjacency_matrix=adj_matrix,
            connectivity_matrix=connectivity_metric_matrix,
            connection_ratio=connection_ratio,
        )

        return info

    @staticmethod
    def _surrogate_t_test(
        source, target, skip_surrogate=False, surrogate_N=30, seed=None
    ):
        """
        Surrogate t-test
        """
        # Function configuration. TODO: Make this dependency injection
        order = 2

        assert source.shape[0] == target.shape[0], (
            f"source.shape={source.shape}, target.shape={target.shape}"
        )

        sig = np.stack([source, target], axis=-1)
        try:
            val = pairwise_granger(sig, order)
        except ValueError:
            val = [0.0, 0.0]
            # val = [np.nan, np.nan]

        surrogate_vals = []
        if skip_surrogate:
            return val, surrogate_vals

        sublength = 64
        stride = 8
        assert source.shape[0] - sublength > 0, (
            f"During surrogate test: source.shape[0]={source.shape[0]}, sublength={sublength}"
        )

        rng = np.random.default_rng(seed)  # TODO take rng instead
        for _ in range(surrogate_N):
            surrogate_source = source.copy()
            rng.shuffle(surrogate_source)
            for start_index in np.arange(0, source.shape[0] - sublength, stride):
                end_index = start_index + sublength
                surr_val = func(
                    np.stack(
                        [
                            surrogate_source[start_index:end_index],
                            target[start_index:end_index],
                        ],
                        axis=-1,
                    ),
                    order,
                )
                surrogate_vals.append(surr_val)

        return val, surrogate_vals

    @staticmethod
    def _get_connection_info(
        pair,
        binned_spiketrain,
        channels,
        mea,
        skip_surrogate,
        surrogate_N,
        seed,
    ):
        """
        Get connection information
        """
        sid, tid = pair
        if sid == tid:
            return 1, [0, 0]
        if mea is not None:
            if channels[sid] not in mea or channels[tid] not in mea:
                return 1, [0, 0]
        source = binned_spiketrain[sid]
        target = binned_spiketrain[tid]
        te, surrogate_te_list = UndirectedConnectivity._surrogate_t_test(
            source,
            target,
            skip_surrogate=skip_surrogate,
            surrogate_N=surrogate_N,
            seed=seed,
        )
        if skip_surrogate:
            return 1, te
        t_value, p_value = spst.ttest_1samp(surrogate_te_list, te, nan_policy="omit")
        return p_value, te

    def plot_adjacency_matrix(self, result, inputs, save_path=None, show=False):
        connectivity_metric_matrix = result["connectivity_matrix"]

        fig, ax = plt.subplots(figsize=(12, 12))
        im = ax.imshow(connectivity_metric_matrix, cmap="gray_r", vmin=0)
        ax.set_xlabel("Target")
        ax.set_ylabel("Source")
        ax.set_title("Transfer Entropy")
        plt.colorbar(im, ax=ax)

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "te.png"))

        if show:
            plt.show()

        plt.close(fig)

    def plot_transfer_entropy_histogram(
        self, result, inputs, save_path=None, show=False
    ):
        connectivity_metric_matrix = result["connectivity_matrix"]

        # Plot values in heatmap
        fig, ax = plt.subplots(figsize=(12, 12))
        ax.hist(connectivity_metric_matrix, bins=30)
        ax.set_xlabel("transfer entropy")
        ax.set_ylabel("count")
        ax.set_title("Transfer Entropy Histogram")

        if save_path is not None:
            plt.savefig(os.path.join(save_path, "te_histogram.png"))

        if show:
            plt.show()

        plt.close(fig)



__all__ = [
    "cache_generator_call",
]

from typing import Any
from collections.abc import Generator, Callable

import time
import functools
import inspect

from .operator import GeneratorOperator


def cache_generator_call(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(
        self: GeneratorOperator,
        idx,
        *args,
    ) -> Generator | Any | None:
        tag = "data"
        cacher = self.cacher
        result = func(self, *args)
        if result is not None:
            cacher.save_cache(result, idx=idx, tag=tag)
            if idx == 0:
                cacher.save_config(tag=tag)  # config saved only once

        # Plotting  # FIXME: Somehow move it in operator mixin?
        self._callback_generator_plot(idx, result, args, save_path=self.analysis_path)
        if idx == 0:
            self._callback_firstiter_plot(result, args, save_path=self.analysis_path)

        return result

    return wrapper


# def cache_generator_call(func: Callable) -> Callable:
#    """
#    Cache the methods of the operator.
#    It is special case for the generator in-out stream.
#    Save the cache in the cacher object with appropriate tag.
#
#    If inputs are not all generators, it will run regular function.
#    """
#
#    def wrapper(
#        self: GeneratorOperator, *args: Any, **kwargs: Any
#    ) -> Generator | Any | None:
#        is_all_generator = all(inspect.isgenerator(v) for v in args) and all(
#            inspect.isgenerator(v) for v in kwargs.values()
#        )
#
#        tag = "data"
#        cacher = self.cacher
#
#        if is_all_generator:
#
#            def generator_func(*args: tuple[Generator, ...]) -> Generator:
#                for idx, zip_arg in enumerate(zip(*args, strict=False)):
#                    stime = time.time()
#                    result = func(self, *zip_arg, **kwargs)
#                    # print(f"    iter {idx:03d} {self}: {time.time() - stime:.03f}sec", flush=True)
#                    if result is not None:
#                        # In case the module does not return anything
#                        cacher.save_cache(result, idx, tag=tag)
#                    self._callback_generator_plot(
#                        idx, result, zip_arg, save_path=self.analysis_path
#                    )
#                    if idx == 0:
#                        self._callback_firstiter_plot(
#                            result, zip_arg, save_path=self.analysis_path
#                        )
#                    yield result
#                cacher.save_config(tag=tag)
#                # TODO: add lastiter_plot
#                # FIXME
#                self._done_flag_generator_plot = True
#                self._done_flag_firstiter_plot = True
#
#            generator = generator_func(*args, *kwargs.values())
#            return generator
#        else:
#            result = func(self, *args, **kwargs)
#            if result is None:
#                # In case the module does not return anything
#                return None
#            cacher.save_cache(result, tag=tag)
#            cacher.save_config(tag=tag)
#            return result
#
#    return wrapper



__doc__ = """ Base mixin for MEA classes. """


import matplotlib.pyplot as plt
import numpy as np

from miv.core.operator.operator import BaseChainingMixin
from miv.core.operator.loggable import DefaultLoggerMixin


class MEAMixin(BaseChainingMixin, DefaultLoggerMixin):
    """Base mixin for MEA classes.

    Functional module.

    Comply with the following interface:
        MEAGeometryProtocol
    """

    def __init__(self, tag: str = "mea", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tag = tag
        self.runner = None

    def get_xy(self, idx: int) -> tuple[float, float]:
        """Given node index, return xy coordinate"""
        raise NotImplementedError

    def save(self, path: str) -> None:
        raise NotImplementedError

    def load(self, path: str) -> None:
        raise NotImplementedError

    def view(self) -> plt.Figure:
        """Simplified view of MEA orientation"""
        raise NotImplementedError

    @property
    def output(self):
        return self

    def run(self, **kwargs):
        return self.output

    def map_data(self, vector: np.ndarray) -> np.ndarray:
        """Map data (1-D array) to MEA (2-D array or N-D array)"""
        raise NotImplementedError

    def set_save_path(self, *args, **kwargs):
        pass



import matplotlib.pyplot


def format_plot2(plt: matplotlib.pyplot):
    plt.style.use("default")
    plt.rcParams["axes.linewidth"] = 2.0
    plt.rcParams["axes.edgecolor"] = "black"
    plt.rcParams["xtick.direction"] = "in"
    plt.rcParams["xtick.major.width"] = 1.5
    plt.rcParams["xtick.minor.width"] = 1.2
    plt.rcParams["ytick.direction"] = "in"
    plt.rcParams["ytick.major.width"] = 1.5
    plt.rcParams["ytick.minor.width"] = 1.2

    # FONT
    SMALL_SIZE = 18
    # MEDIUM_SIZE = 21
    BIGGER_SIZE = 24
    plt.rcParams["font.family"] = "Times New Roman"
    plt.rcParams["font.size"] = 24
    plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
    plt.rc("axes", titlesize=BIGGER_SIZE)  # fontsize of the axes title
    plt.rc("axes", labelsize=BIGGER_SIZE)  # fontsize of the x and y labels
    plt.rc("xtick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
    plt.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
    plt.rc("legend", fontsize=SMALL_SIZE)  # legend fontsize
    plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title



from collections.abc import Sequence
import numpy as np


def downsample_average(
    x: Sequence[float],
    y: Sequence[float],
    max_samples: int = 64,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Downsample x and y by averaging over non-overlapping windows so that
    the returned length is approximately max_samples.

    If len(x) <= max_samples, returns the original data as numpy arrays.

    Args:
        x: Sequence of numeric values (e.g. list or 1D array).
        y: Sequence of numeric values, same length as x.
        max_samples: Desired maximum number of output samples.

    Returns:
        A tuple (x_ds, y_ds) of numpy arrays, each of length ~max_samples.
    """
    arr_x = np.asarray(x)
    arr_y = np.asarray(y)
    size = arr_x.shape[0]
    if size <= max_samples:
        return arr_x, arr_y

    best_n = size // max_samples

    # Number of full windows
    bins = size // best_n

    # Trim off any leftover at the end
    x_trim = arr_x[: bins * best_n].reshape(bins, best_n)
    y_trim = arr_y[: bins * best_n].reshape(bins, best_n)

    # Compute the average in each window
    x_ds = x_trim.mean(axis=1)
    y_ds = y_trim.mean(axis=1)
    return x_ds, y_ds



import logging

import click

from miv.core.pipeline import Pipeline
from miv.signal.filter import ButterBandpass
from miv.signal.spike import ThresholdCutoff


@click.command()
@click.option("--path", "-p", help="Path to the recorded directory.", multiple=True)
@click.option(
    "--tools",
    "-t",
    type=click.Choice(["OpenEphys", "Intan"]),
    default="OpenEphys",
    help="Select the recording format",
)
@click.option(
    "--nproc",
    "-N",
    default=None,
    help="Number of processing units. If not provided, use maximum.",
)
@click.option(
    "--num_fragments",
    "-n",
    default=1,
    help="Number of fragments for data processing. Recommend to split data into 1 minute recording segments. For Intan, this number is pre-determined by the number of saved files.",
)
@click.option(
    "--use-mpi",
    type=bool,
    default=False,
    help="Set True if mpi is ready. Else, it will use multiprocessing. (mpi4py must be installed)",
)
@click.option("--chunksize", default=1, help="Number of chunks for multiprocessing")
def main(
    path: str, tools: str, nproc: int, num_fragments: int, use_mpi: bool, chunksize: int
):
    signal_filter = ButterBandpass(400, 1500, order=4)
    spike_detection = ThresholdCutoff(cutoff=5.0, progress_bar=True)
    signal_filter >> spike_detection

    # Select tools
    if tools == "OpenEphys":
        # If OpenEphys is selected, process all recordings in the set.
        from miv.io.openephys import DataManager

        for p in path:
            for data in DataManager(p):
                data >> signal_filter
                pipeline = Pipeline(spike_detection)
                pipeline.run(data.analysis_path)
                spike_detection.plot(save_path=data.analysis_path)
                logging.info(f"Pre-processing {p}-{data.data_path} done.")
                data.clear_connections()
    elif tools == "Intan":
        from miv.io.intan import DataIntan

        for p in path:
            data = DataIntan(p)
            pipeline = Pipeline(spike_detection)
            pipeline.run(data.analysis_path)
            spike_detection.plot(save_path=data.analysis_path)
            logging.info(f"Pre-processing {p}-{data.data_path} done.")
            data.clear_connections()


if __name__ == "__main__":
    main()



from __future__ import annotations

__doc__ = """

-------------------------------------

Raw Data Loader
###############

"""
__all__ = [
    "load_continuous_data",
    "load_recording",
    "oebin_read",
    "apply_channel_mask",
    "load_ttl_event",
]

import logging
import math
import os
from ast import literal_eval
from collections.abc import Sequence
from glob import glob
from typing import TYPE_CHECKING, Any

import numpy as np
import quantities as pq

from miv.typing import SignalType

if TYPE_CHECKING:
    pass

try:
    from mpi4py import MPI

    comm = MPI.COMM_WORLD
    logger = logging.getLogger(f"rank[{comm.Get_rank()}]-OpenEphys")
except ImportError:
    logger = logging.getLogger(__name__)


def apply_channel_mask(signal: np.ndarray, channel_mask: set[int]):
    """Apply channel mask on the given signal.

    Parameters
    ----------
    signal : np.ndarray
        Shape of the signal is expected to be (num_data_point, num_channels).
    channel_mask : Set[int]

    Returns
    -------
    output signal : SignalType

    Raises
    ------
    IndexError
        Typically raise index error when the dimension of the signal is less than 2.
    AttributeError
        If signal is non numpy array type.

    """
    num_channels = signal.shape[1]
    channel_index_set = set(range(num_channels)) - channel_mask
    channel_index = np.array(np.sort(list(channel_index_set)))
    signal = signal[:, channel_index]
    return signal


def bits_to_voltage(signal: SignalType, channel_info: Sequence[dict[str, Any]]):
    """
    Convert binary bit data to voltage (microVolts)

    Parameters
    ----------
    signal : SignalType, numpy array
    channel_info : Dict[str, Dict[str, Any]]
        Channel information dictionary. Typically located in `structure.oebin` file.
        Channel information includes bit-volts conversion ration and units (uV or mV).

    Returns
    -------
    signal : numpy array
        Output signal is in microVolts (uV) unit.

    """
    resultant_unit = pq.Quantity(1, "uV")  # Final Unit
    for channel in range(len(channel_info)):
        bit_to_volt_conversion = channel_info[channel]["bit_volts"]
        recorded_unit = pq.Quantity([1], channel_info[channel]["units"])
        unit_conversion = (recorded_unit / resultant_unit).simplified
        signal[:, channel] *= bit_to_volt_conversion * unit_conversion
    return signal


def oebin_read(file_path: str):
    """
    Oebin file reader in dictionary form

    Parameters
    ----------
    file_path : str

    Returns
    -------
    info : Dict[str, any]
        recording information stored in oebin file.
    """
    # TODO: may need fix for multiple continuous data.
    # TODO: may need to include processor name/id
    info = literal_eval(open(file_path).read())
    return info


def load_ttl_event(
    folder: str,
    return_sample_numbers: bool = False,
):
    """
    Loads TTL event data recorded by Open Ephys as numpy arrays.

    `Reference: OpenEphys TTL data structure <https://open-ephys.github.io/gui-docs/User-Manual/Recording-data/Binary-format.html#events>`_

    The path should contain:

    - states.npy: N 16-bit integers, indicating ON/OFF (channel number)
    - sample_numbers.npy: N 64-bit integers, sample number during acquisition.
    - timestamps.npy: N 64-bit floats, global timestamps
    - full_words.npy: N 64-bit integer, TTL word of current state of all lines.

    Extra data are retrieved from:

    - structure.oebin: number of channels and sampling rate.

    Parameters
    ----------
    folder: str
        Folder containing the subfolder 'experiment1'.
    return_sample_numbers: bool
        If set to true, also return sample_numbers that can be used to re-calculate
        synchronization between time series. (default=False)

    Returns
    -------
    states : np.ndarray
        Numpy integer array, indicating ON/OFF state. (+- channel number)
    full_words : np.ndarray
        Numpy integer array, consisting current state of all lines.
    timestamps : np.ndarray
        Numpy float array. Global timestamps in seconds. Relative to start
        of the Record Node's main data stream.
    sampling_rate: float
        Recorded sampling rate
    initial_state: int
        Initial TTL state across lines.
    sample_numbers: Optional[np.ndarray]
        Return if `return_sample_numbers` is true. Return array of sample numbers that
        records sampled clock count. Typically used to synchronize time array.

    Raises
    ------
    AssertionError
        No events recorded in data.

    """
    # Check TTL event recorded
    info_file: str = os.path.join(folder, "structure.oebin")
    info: dict[str, Any] = oebin_read(info_file)
    version = info["GUI version"]
    assert "events" in info.keys(), "No events recorded (TTL)."
    ttl_info = [
        data
        for data in info["events"]
        if "TTL Input" in data["channel_name"] or "TTL events" in data["channel_name"]
    ]
    assert len(ttl_info) > 0, "No events recorded (TTL)."
    assert len(ttl_info) == 1, (
        "Multiple TTL input is found, which is not supported yet. (TODO)"
    )
    ttl_info = ttl_info[0]

    # Data Structure (OpenEphys Structure)
    v_major, v_minor, v_sub = map(int, version.split("."))
    if v_major == 0 and v_minor <= 5 and v_sub == 4:  # Legacy file name before 0.6.0
        file_states = "channel_states.npy"
        file_timestamps = "timestamps.npy"
        file_sample_numbers = "channels.npy"
        file_full_words = "full_words.npy"
    elif v_major == 0 and v_minor <= 5:  # Legacy file name before 0.6.0
        file_states = "states.npy"
        file_timestamps = "synchronized_timestamps.npy"
        file_sample_numbers = "timestamps.npy"
        file_full_words = "full_words.npy"
    else:
        file_states = "states.npy"
        file_timestamps = "timestamps.npy"
        file_sample_numbers = "sample_numbers.npy"
        file_full_words = "full_words.npy"
    file_path = os.path.join(folder, "events", ttl_info["folder_name"])

    states = np.load(os.path.join(file_path, file_states)).astype(np.int16)
    sample_numbers = np.load(os.path.join(file_path, file_sample_numbers)).astype(
        np.int64
    )
    timestamps = np.load(os.path.join(file_path, file_timestamps)).astype(np.float64)
    full_words = np.load(os.path.join(file_path, file_full_words)).astype(np.int64)

    # Load from structure.oebin file
    sampling_rate: float = ttl_info["sample_rate"]
    initial_state: int = ttl_info["initial_state"] if "initial_state" in ttl_info else 0

    if return_sample_numbers:
        return (
            states,
            full_words,
            timestamps,
            sampling_rate,
            initial_state,
            sample_numbers,
        )
    else:
        return states, full_words, timestamps, sampling_rate, initial_state


def load_recording(
    folder: str,
    channel_mask: set[int] | None = None,
    start_at_zero: bool = True,
    dtype: np.dtype = np.float32,
    progress_bar: bool = False,
    mpi_comm: mpi4py.MPI.Comm | None = None,
    _recorded_dtype="int16",
):
    """
    Loads data recorded by Open Ephys in Binary format as numpy memmap.
    The path should contain

    - continuous/<processor name>/continuous.dat: signal (cannot have multiple file)
    - continuous/<processor name>/timestamps.npy: timestamps
    - structure.oebin: number of channels and sampling rate.

    Parameters
    ----------
    folder: str
        folder containing at least the subfolder 'experiment1'.
    channel_mask: Set[int], optional
        Channel index list to ignore in import (default=None)
    start_at_zero : bool
        If True, the timestamps is adjusted to start at zero.
        Note, recorded timestamps might not start at zero for some reason.
    num_fragments : Optional[int]
        Instead of loading entire data at once, split the data into `num_fragment`
        number of subdata to process separately. By default, num_fragments is set
        to split the recording into 1 minute segments. (default=None)
    start_index : Optional[int]
        Start index of the fragments. It is useful when you want to submit MPI processing.
        For example, one can submit num_fragments=10, start_index=3 to process 3-9 fragments.
        (Zero-indexing)
    end_index : Optional[int]
        End index of the fragments. It is useful when you want to submit MPI processing.
        For example, one can submit num_fragments=10, end_index=5 to process 0-5 fragments.
        (Zero-indexing)
    dtype: np.dtype
        If None, skip data-type conversion. If the filesize is too large, it is advisable
        to keep `dtype=None` and convert slice by slice. (default=float32)

    Returns
    -------
    signal : SignalType, neo.core.AnalogSignal
    timestamps : np.ndarray
    sampling_rate : float

    Raises
    ------
    AssertionError
        If more than one "continuous.dat" file exist in the directory.

    """
    file_path: list[str] = glob(
        os.path.join(folder, "**", "continuous.dat"), recursive=True
    )
    assert len(file_path) == 1, (
        f"There should be only one 'continuous.dat' file. (There exists {file_path})"
    )

    # load structure information dictionary
    info_file: str = os.path.join(folder, "structure.oebin")
    info: dict[str, Any] = oebin_read(info_file)
    num_channels: int = info["continuous"][0]["num_channels"]
    sampling_rate: int = int(info["continuous"][0]["sample_rate"])
    # channel_info: Dict[str, Any] = info["continuous"][0]["channels"]

    _old_oe_version = False
    if "GUI version" in info:
        version = info["GUI version"]
        v_major, v_minor, v_sub = list(map(int, version.split(".")))[:3]
        _old_oe_version = v_major == 0 and v_minor <= 5  # Legacy

    # Read timestamps first
    dirname = os.path.dirname(file_path[0])
    timestamps_path = os.path.join(dirname, "timestamps.npy")
    timestamps = load_timestamps(timestamps_path, sampling_rate, _old_oe_version)
    total_length = timestamps.size

    # Define task
    filesize = os.path.getsize(file_path[0])
    itemsize = np.dtype(_recorded_dtype).itemsize
    if filesize != itemsize * total_length * num_channels:
        logger.warning(
            f"{filesize=} does not match the expected {itemsize*total_length*num_channels=}.\n"
            f"Possibly a file is corrupted or recording crashed."
            f"Path: {file_path[0]}"
        )
        total_length = int(np.floor(filesize / itemsize * num_channels))
        timestamps = timestamps[:total_length]
    samples_per_block = sampling_rate * 60  # TODO: 1 minute chunking
    num_fragments = int(math.ceil(total_length / samples_per_block))
    tasks = None
    if mpi_comm is not None:
        # ex) split [1,2,3,4] --> [1,2], [3,4]
        from miv.utils.mpi import task_index_split

        tasks = task_index_split(mpi_comm, num_fragments)
    else:
        # None-mpi case: Load all data and parse
        tasks = list(range(num_fragments))
    logger.info(f"tasks: {tasks}")
    if len(tasks) == 0:
        logger.warning("Number of processor(size) exceeded the available tasks.")
        return

    # Data readout
    offset = itemsize * num_channels * tasks[0] * samples_per_block
    if tasks[-1] == num_fragments - 1:  # include last block
        shape = None  # Read rest
    else:
        shape = (len(tasks) * samples_per_block, num_channels)
    signal = load_continuous_data(
        file_path[0], num_channels, _recorded_dtype, offset, shape
    )
    logger.info(f"Memory read: {signal.shape=}")

    # Output
    for i, task in enumerate(tasks):
        _signal = signal[
            i * samples_per_block : min((i + 1) * samples_per_block, signal.shape[0])
        ]
        _timestamps = timestamps[
            task * samples_per_block : min((task + 1) * samples_per_block, total_length)
        ]
        logger.info(
            f"generate {_signal.shape=}, {_timestamps.shape=}, {sampling_rate=}"
        )
        logger.info(
            f"signal: {i * samples_per_block} - {min((i + 1) * samples_per_block, signal.shape[0])}"
        )
        logger.info(
            f"timestamps: {task * samples_per_block} - {min((task + 1) * samples_per_block, total_length)}"
        )

        _signal = _signal.astype(dtype)
        logger.info("Array cast done")

        # To Voltage
        _signal = bits_to_voltage(_signal, info["continuous"][0]["channels"])

        if channel_mask:
            _signal = apply_channel_mask(_signal, channel_mask)

        # Adjust timestamps to start from zero
        if start_at_zero and not np.isclose(_timestamps[0], 0.0):
            _timestamps -= _timestamps[0]

        if _timestamps.size / sampling_rate < 0.001:
            break
        yield _signal, _timestamps, sampling_rate


def load_continuous_data(
    data_path: str,
    num_channels: int,
    _recorded_dtype: np.dtype | str = "int16",
    offset: int = 0,
    shape: tuple[int, int] = None,
):
    """
    Load single continous data file and return timestamps and raw data in numpy array.
    Typical `data_path` from OpenEphys has a name `continuous.dat`.

    .. note::
        The output data is raw-data without unit conversion. In order to convert the unit
        to voltage, you need to multiply by `bit_volts` conversion ratio. This ratio and
        units are typially saved in `structure.oebin` file.

    Parameters
    ----------
    data_path : str
        continuous.dat file path from Open_Ethys recording.
    num_channels : int
        number of recording channels recorded. Note, this method will not throw an error
        if you don't provide the correct number of channels.
    _recorded_dtype: Union[np.dtype, str]
        Recorded data type.

    Returns
    -------
    raw_data: SignalType, numpy array
    timestamps: np.ndarray

    Raises
    ------
    FileNotFoundError
        If data_path is invalid.
    ValueError
        If the error message shows the array cannot be reshaped due to shape,
        make sure the num_channels is set accurately.

    """
    # Read raw data signal
    if os.path.getsize(data_path) == 0:
        raise ValueError(f"Data {data_path} has zero-size.")
    raw_data: np.ndarray = np.memmap(
        data_path,
        dtype=_recorded_dtype,
        mode="r",
        offset=offset,
        shape=shape,
        order="C",
    )
    if shape is None:
        length = raw_data.size // num_channels
        raw_data = raw_data.reshape(length, num_channels)

    return raw_data


def load_timestamps(
    timestamps_path: str,
    sampling_rate: float,
    _old_oe_version: bool = False,
):
    """
    Load single continous data file and return timestamps and raw data in numpy array.
    Typical `data_path` from OpenEphys has a name `continuous.dat`.

    .. note::
        The output data is raw-data without unit conversion. In order to convert the unit
        to voltage, you need to multiply by `bit_volts` conversion ratio. This ratio and
        units are typially saved in `structure.oebin` file.

    Parameters
    ----------
    timestamps_path : str
        Path for the timestamps file.
    sampling_rate : float
        data sampling rate.

    Returns
    -------
    raw_data: SignalType, numpy array
    timestamps: np.ndarray

    Raises
    ------
    FileNotFoundError
        If data_path is invalid.
    ValueError
        If the error message shows the array cannot be reshaped due to shape,
        make sure the num_channels is set accurately.

    """
    # Get timestamps
    timestamps = np.asarray(
        np.load(timestamps_path)
    )  # TODO: check if npy file includes dtype. else, add "dtype=np.float32"
    if _old_oe_version:
        timestamps = timestamps / float(sampling_rate)

    return timestamps



__doc__ = """
Export prb files, which is inherited from phy.
"""

import os
import pathlib

import click
import numpy as np
from miv_os_contrib.impedance_check import (
    get_channel_in_impedance_range,
    get_impedances_from_csv,
)

from miv.mea import mea_map
from miv.mea.channel_mapping import MEA64, MEA128


@click.command()
@click.option("--mea-name", "-m", help="Name of the mea.")
@click.option("--spacing", "-s", type=int, help="Path to the impedance file.")
@click.option(
    "--output", "-o", type=click.Path(exists=False), help="Path to the output file."
)
@click.option(
    "--impedance-path",
    "-i",
    default=None,
    type=click.Path(),
    help="Path to the impedance file.",
)
@click.option(
    "--is-intan",
    "-intan",
    type=bool,
    default=False,
    is_flag=True,
    help="Whether the file is recorded from intan.",
)
@click.option(
    "--total-nb-channels",
    "-t",
    type=int,
    default=64,
    help="Total number of channels. Default is 64.",
)
def export(
    mea_name, spacing, output, impedance_path, is_intan: bool, total_nb_channels: int
):
    probes_path = pathlib.Path(os.environ["HOME"]) / "spyking-circus" / "probes"

    if is_intan:
        if mea_name == "64_intanRHD":
            grid = MEA64(mea_name).mea_intan
        elif mea_name == "128_dual_connector_two_64_rhd":
            grid = MEA128(mea_name).mea_intan
    else:
        grid = mea_map[mea_name]

    if impedance_path is not None:
        range_impedance = (1e5, 1.3e6)
        low_impedance, high_impedance = range_impedance
        if impedance_path.endswith(".xml"):
            # Impedances
            impedances = get_channel_in_impedance_range(impedance_path, range_impedance)
            channels_with_impedances_in_range = list(impedances.keys())
        elif impedance_path.endswith(".csv"):
            # Impedances
            impedances = get_impedances_from_csv(impedance_path)
            channels_with_impedances_in_range = np.where(
                np.logical_and(
                    impedances > low_impedance,
                    impedances < high_impedance,
                )
            )[0]
        print("Channels with impedances in range:")
        for ch in channels_with_impedances_in_range:
            print(f"Channel {ch} has impedance {impedances[ch]:.2f}")
    else:
        channels_with_impedances_in_range = np.arange(total_nb_channels)

    # Locations
    locs = {}
    for ch in channels_with_impedances_in_range:
        where = np.where(grid == ch)
        y, x = where[0][0], where[1][0]
        locs[ch] = (x * spacing, y * spacing)

    text = configuration_to_string(
        locs, channels_with_impedances_in_range, total_nb_channels=total_nb_channels
    )

    with open(probes_path / output, "w") as f:
        f.write(text)


def configuration_to_string(locs, channels, radius=250, total_nb_channels=64):
    geometry = {ch: list(locs[ch]) for ch in channels}
    channel_groups = {
        1: {"channels": list(channels), "graph": [], "geometry": geometry}
    }

    return f"""total_nb_channels = {total_nb_channels}
radius            = {radius}

channel_groups = {channel_groups}
"""


if __name__ == "__main__":
    export()



__doc__ = """
Spikestamps
===========

.. autoclass:: Spikestamps
   :members:

"""

__all__ = ["Spikestamps"]

from collections.abc import MutableSequence, Sequence, Iterable

import numpy as np
import quantities as pq

import neo

from .mixin_colapsable import ConcatenateMixin
from .signal import Signal
from miv.core.operator.operator import DataNodeMixin


class Spikestamps(ConcatenateMixin, DataNodeMixin, Sequence):
    """List of array of spike times

    Represents spikes emitted by the same unit in a period of times.

    Comply with `ChannelWise` and `Extendable` protocols.
    """

    def __init__(self, iterable: list | None = None) -> None:
        super().__init__()
        if iterable is None:  # Default
            iterable = []
        self.data: list[MutableSequence[float]] = iterable

    @classmethod
    def empty(clf, number_of_channels: int) -> "Spikestamps":
        return clf([[] for _ in range(number_of_channels)])

    @property
    def number_of_channels(self) -> int:
        """Number of channels"""
        return len(self.data)

    def __setitem__(self, index: int, item: MutableSequence[float]) -> None:
        self.data[index] = item

    def __getitem__(self, index: int) -> MutableSequence[float]:  # type: ignore[override]
        return self.data[index]

    def __len__(self) -> int:
        return len(self.data)

    def insert(self, index: int, item: MutableSequence[float]) -> None:
        if index > len(self.data) or index < 0:
            raise IndexError("Index out of range")
        self.data.insert(index, item)

    def append(self, item: MutableSequence[float]) -> None:
        self.data.append(item)

    def extend(self, other: "Spikestamps | Iterable[MutableSequence[float]]") -> None:
        """
        Extend spikestamps from another `Spikestamps` or list of arrays.

        If the given parameter is another `Spikestamps`, each arrays are concatenated.
        Remaining channels will be added as different channels.

        If the given parameter is list of another arrays, each arrays will be added as different channels.

        Examples
        --------
        >>> a = Spikestamps([[0,1,2],[0,3]])
        >>> b = Spikestamps([[3,5],[4],[0,1]])
        >>> a.extend(b)
        >>> a
        [[0,1,2,3,5], [0,3,4], [0,1]]

        >>> c = [[1],[5],[0]]
        >>> a.extend(c)
        >>> a
        [[0,1,2,3,5], [0,3,4], [0,1], [1], [5], [0]]

        """
        if isinstance(other, type(self)):
            length_diff = len(other) - len(self.data)
            if length_diff > 0:
                for _ in range(length_diff):
                    self.data.append([])
            for idx, array in enumerate(other):
                self.data[idx].extend(array)
        else:
            self.data.extend(item for item in other)

    def get_count(self) -> list[int]:
        """Return list of spike-counts for each channel."""
        return [len(data) for data in self.data]

    def get_last_spikestamp(self) -> float:
        """Return timestamps of the last spike in this spikestamps"""
        rowmax = [max(data) for data in self.data if len(data) > 0]
        return 0 if len(rowmax) == 0 else max(rowmax)

    def get_first_spikestamp(self) -> float:
        """Return timestamps of the first spike in this spikestamps"""
        rowmin = [min(data) for data in self.data if len(data) > 0]
        return 0 if len(rowmin) == 0 else min(rowmin)

    def get_view(
        self, t_start: float, t_end: float, reset_start: bool = False
    ) -> "Spikestamps":
        """
        Truncate array and only includes spikestamps between t_start and t_end.
        If reset_start is True, the first spikestamp will be set to zero.
        """
        spikestamps_array = [
            np.array(sorted(filter(lambda x: t_start <= x <= t_end, arr)))
            for arr in self.data
        ]
        if reset_start:
            spikestamps_array = [
                (arr - t_start) if arr.size > 0 else arr for arr in spikestamps_array
            ]
        return Spikestamps(spikestamps_array)

    def select(self, indices: Sequence[int], keepdims: bool = True) -> "Spikestamps":
        """Select channels by indices. The order of the channels will be preserved."""
        if keepdims:
            data = [
                self.data[idx] if idx in indices else []
                for idx in range(self.number_of_channels)
            ]
            return Spikestamps(data)
        else:
            return Spikestamps([self.data[idx] for idx in indices])

    def neo(
        self, t_start: float | None = None, t_stop: float | None = None
    ) -> list[neo.SpikeTrain]:
        """Cast to neo.SpikeTrain

        Parameters
        ----------
        t_start : float
            Start time (in second) of the spike train. (default=None)
            If None, the first spikestamp will be used.
        t_stop : float
            End time (in second) of the spike train. (default=None)
            If None, the last spikestamp will be used.

        """
        if t_start is None:
            t_start = self.get_first_spikestamp()
        if t_stop is None:
            t_stop = self.get_last_spikestamp()
        return [
            neo.SpikeTrain(arr, t_start=t_start, t_stop=t_stop, units=pq.s)
            for arr in self.data
        ]

    def flatten(self) -> tuple[np.ndarray, np.ndarray]:
        """Flatten spikestamps into a single array. One can plot the spikestamps using this array with scatter plot."""
        x: list[float] = []
        y: list[float] = []
        for idx, arr in enumerate(self.data):
            x.extend(arr)
            y.extend([idx] * len(arr))
        return np.array(x), np.array(y)

    def binning(
        self,
        bin_size: float = 0.001,
        t_start: float | None = None,
        t_end: float | None = None,
        minimum_count: int = 1,
        return_count: bool = False,
    ) -> Signal:
        """
        Forms a binned spiketrain using the spiketrain

        Parameters
        ----------
        bin_size : float | pq.Quantity
            bin size in the unit of time.
        t_start : float
            start time of the binned spiketrain. (default=None)
        t_end : float
            end time of the binned spiketrain. (default=None)
        minimum_count : int
            Minimum number of spikes in a bin to be considered as a spike. (default=1)
            If return_count is set to True, this parameter is ignored.
        return_count : bool
            If set to true, return the bin count. (default=False)

        Returns
        -------
        bin_spike: numpy.ndarray
            binned spiketrain with 1 corresponding to spike and zero otherwise

        """
        if isinstance(bin_size, pq.Quantity):
            bin_size = bin_size.rescale(pq.s).magnitude
        assert bin_size > 0, "bin size should be greater than 0"
        if isinstance(t_start, pq.Quantity):
            t_start = t_start.rescale(pq.s).magnitude
        if isinstance(t_end, pq.Quantity):
            t_end = t_end.rescale(pq.s).magnitude

        t_start = self.get_first_spikestamp() if t_start is None else t_start
        t_end = self.get_last_spikestamp() if t_end is None else t_end
        assert t_start < t_end, (
            f"t_start ({t_start}) should be less than t_end ({t_end})"
        )
        n_bins = int(np.ceil((t_end - t_start) / bin_size))
        time = t_start + (np.arange(n_bins + 1) * bin_size)

        num_channels = self.number_of_channels
        signal = Signal(
            data=np.zeros(
                [time.shape[0] - 1, num_channels],
                dtype=np.int_ if return_count else np.bool_,
            ),
            timestamps=time[:-1],
            rate=1.0 / bin_size,
        )
        for idx, spiketrain in enumerate(self.data):
            bins = np.digitize(np.asarray(spiketrain), time)
            bincount = np.bincount(bins, minlength=n_bins + 2)[1:-1]
            if return_count:
                bin_spike = bincount
            else:
                bin_spike = (bincount >= minimum_count).astype(np.bool_)
            signal.data[:, idx] = bin_spike
        return signal

    def get_portion(self, start_ratio: float, end_ratio: float) -> "Spikestamps":
        """
        (Experimental)
        Return spiketrain view inbetween (start_ratio, end_ratio)
        where ratio=0 indicates the first spikestamps, and ratio=1 represent last spikestamp.
        """
        t_start = self.get_first_spikestamp()
        t_end = self.get_last_spikestamp()
        return self.get_view(
            t_start + (t_end - t_start) * start_ratio,
            t_start + (t_end - t_start) * end_ratio,
        )

    @classmethod
    def from_pickle(cls, filename: str) -> "Spikestamps":
        import pickle as pkl

        with open(filename, "rb") as f:
            data = pkl.load(f)

        return cls(data)

    def shift(self, shift: float) -> "Spikestamps":
        """Shift all spikestamps by a constant value

        Parameters
        ----------
        shift : float
            Value to shift the spikestamps.
        """
        return Spikestamps([np.asarray(arr) + shift for arr in self.data])



from __future__ import annotations

__all__ = [
    "GeneratorOperatorMixin",
]

from typing import TYPE_CHECKING, Protocol
from collections.abc import Generator


if TYPE_CHECKING:
    pass

from miv.core.operator.cachable import CACHE_POLICY
from miv.core.operator.operator import OperatorMixin
from miv.core.operator.protocol import _Node
from miv.core.operator_generator.callback import (
    GeneratorCallbackMixin,
    _GeneratorCallback,
)
from miv.core.operator_generator.policy import VanillaGeneratorRunner


class GeneratorOperator(
    _Node,
    _GeneratorCallback,
    Protocol,
): ...


class GeneratorOperatorMixin(OperatorMixin, GeneratorCallbackMixin):
    def __init__(self) -> None:
        super().__init__()

        self.runner = VanillaGeneratorRunner(self)

    def set_caching_policy(self, policy: CACHE_POLICY) -> None:
        self.cacher.policy = policy

    def output(self) -> Generator:
        """
        Output viewer. If cache exist, read result from cache value.
        Otherwise, execute (__call__) the module and return the value.
        """
        if self.cacher.check_cached():
            self.logger.info(f"Using cache: {self.cacher.cache_dir}")

            def generator_func() -> Generator:
                yield from self.cacher.load_cached()

            output = generator_func()
        else:
            self.logger.info("Cache not found.")
            self.logger.info(f"Using runner: {self.runner.__class__} type.")
            args = self.receive()  # Receive data from upstream
            assert len(args) > 0, (
                "No data received from upstream. Generator-operator must receive other generators from upstream."
            )
            output = self.runner(self.__call__, args)

            # Callback: After-run
            self._callback_after_run(output)

            # Plotting: Only happened when cache is not called
            self._callback_plot(output, args, show=False)
        return output



from typing import Protocol
from collections.abc import Iterable
from functools import reduce


class _Collapsable(Protocol):
    """A protocol defining the concatenate method for collapsable objects."""

    @classmethod
    def concatenate(
        cls, values: Iterable["_Collapsable"], *, head: "_Collapsable | None" = None
    ) -> "_Collapsable": ...

    def extend(self, other: "_Collapsable") -> "_Collapsable": ...


class ConcatenateMixin:
    """A mixin providing concatenate functionality for collapsable objects.
    Allows combining multiple objects by extending them sequentially."""

    @classmethod
    def concatenate(
        cls, values: Iterable[_Collapsable], *, head: "_Collapsable | None" = None
    ) -> _Collapsable:
        """
        Concatenate a list of collapsable objects.
        Note, the method change the data in-place. The concatenated result will be
        stacked on the first object. if `head` is provided, the concatenated result will be
        stacked on the `head`.

        Args:
            values: A list of collapsable objects to concatenate.
            head: An optional collapsable object to use as the head of the concatenation.

        Returns:
            A new collapsable object that is the result of concatenating the input objects.
            The original objects are not modified.

        """
        if head is None:
            return reduce(lambda x, y: x.extend(y), values)
        else:
            return reduce(lambda x, y: x.extend(y), values, head)

    # Deprecated alias
    from_collapse = concatenate



from miv.signal.spike.detection import *
from miv.signal.spike.protocol import *
from miv.signal.spike.sorting import *
from miv.signal.spike.waveform import *
from miv.signal.spike.waveform_statistical_filter import *



__doc__ = """
Implementation for callback features that will be mixed in operator class.
"""
__all__ = ["BaseCallbackMixin"]

from typing import Any
from collections.abc import Callable
from typing_extensions import Self

import types
import inspect
import os
import pathlib

import matplotlib.pyplot as plt

from miv.core.operator.cachable import (
    _CacherProtocol,
    CACHE_POLICY,
)


# MixinOperators
def get_methods_from_feature_classes_by_startswith_str(
    cls: Any, method_name: str
) -> list[Callable]:
    methods = [
        getattr(cls, k)
        for k in dir(cls)
        if k.startswith(method_name) and callable(getattr(cls, k))
    ]
    return methods


# MixinOperators
def get_methods_from_feature_classes_by_endswith_str(
    cls: Any, method_name: str
) -> list[Callable]:
    methods = [
        getattr(cls, k)
        for k in dir(cls)
        if k.endswith(method_name) and callable(getattr(cls, k))
    ]
    return methods


class BaseCallbackMixin:
    def __init__(
        self,
        *args: Any,
        cacher: _CacherProtocol,
        cache_path: str = ".cache",
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.__cache_directory_name: str = cache_path
        self._cacher: _CacherProtocol = cacher

        # Default analysis path
        assert self.tag != "", (
            "All operator must have self.tag attribute for identification."
        )
        self.set_save_path("results")  # FIXME

        # Callback Flags (to avoid duplicated run)
        self._done_flag_after_run = False
        self._done_flag_plot = False

        # Attribute from upstream
        self.tag: str

    @property
    def cacher(self) -> _CacherProtocol:
        return self._cacher

    @cacher.setter
    def cacher(self, value: _CacherProtocol) -> None:
        # FIXME:
        policy = self._cacher.policy
        cache_dir = self._cacher.cache_dir
        self._cacher = value
        self._cacher.policy = policy
        self._cacher.cache_dir = cache_dir

    def set_caching_policy(self, policy: CACHE_POLICY) -> None:
        self.cacher.policy = policy

    def reset_callbacks(self, *, after_run: bool = False, plot: bool = False) -> None:
        self._done_flag_after_run = after_run
        self._done_flag_plot = plot

    def __lshift__(self, right: Callable) -> Self:
        # Dynamically add new function into an operator instance
        if inspect.getfullargspec(right)[0][0] == "self":
            setattr(self, right.__name__, types.MethodType(right, self))
        else:
            # Add new function into as attribute
            setattr(self, right.__name__, right)
        return self

    def set_save_path(
        self,
        path: str | pathlib.Path,
        cache_path: str | pathlib.Path | None = None,
    ) -> None:
        if cache_path is None:
            cache_path = path

        # Set analysis path
        self.analysis_path = os.path.join(path, self.tag.replace(" ", "_"))
        # Set cache path
        _cache_path = os.path.join(
            cache_path, self.tag.replace(" ", "_"), self.__cache_directory_name
        )
        self.cacher.cache_dir = _cache_path

        # Make directory  # Not sure if this needs to be done here
        os.makedirs(self.analysis_path, exist_ok=True)

    def _callback_after_run(self, *args: Any, **kwargs: Any) -> None:
        if self._done_flag_after_run:
            return

        predefined_callbacks = get_methods_from_feature_classes_by_startswith_str(
            self, "after_run"
        )
        for callback in predefined_callbacks:
            callback(*args, **kwargs)

        self._done_flag_after_run = True

    def _callback_plot(
        self,
        output: Any | None,
        inputs: list | None = None,
        show: bool = False,
        save_path: str | pathlib.Path | None = None,
    ) -> None:
        """
        Run all function in this operator that starts with the name 'plot_'.
        """
        if self._done_flag_plot:
            return

        if save_path is None:
            save_path = self.analysis_path

        # If input is single-argument, strip the list
        if isinstance(inputs, list) and len(inputs) == 1:
            inputs = inputs[0]

        plotters = get_methods_from_feature_classes_by_startswith_str(self, "plot_")
        for plotter in plotters:
            plotter(output, inputs, show=show, save_path=save_path)
        if not show:
            plt.close("all")

        self._done_flag_plot = True



__all__ = [
    "probability_distribution",
    "shannon_entropy",
    "block_entropy",
    "entropy_rate",
    "active_information",
    "mutual_information",
    "relative_entropy",
    "joint_entropy",
    "conditional_entropy",
    "cross_entropy",
    "transfer_entropy",
    "partial_information_decomposition",
]


from typing import Literal
from collections.abc import Callable


import numpy as np
from tqdm import tqdm

from miv.core.datatype import Spikestamps
from miv.statistics.spiketrain_statistics import binned_spiketrain

INFO_METRICS_TAG = Literal["self", "pair", "all"]  # Pragma: no cover


def tag_info_metrics(tag: INFO_METRICS_TAG) -> Callable:  # Pragma: no cover
    """
    Decorator to tag the info metrics functions
        - self: channel-wise metrics
        - pair: channel pair-wise metrics
        - all: combined across all channels
    """

    def decorator(func: Callable) -> Callable:
        func.tag = tag
        return func

    return decorator


@tag_info_metrics("self")
def probability_distribution(
    spiketrains: Spikestamps,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Forms the probability distribution required to compute the information theory measures. Probability is computed based on the binned spiketrain generated for the specified bin size.

    Parameters
    ----------
    spiketrains : Spikestamps
    bin_size : float
        bin size in seconds
    t_start : Optional[float]
        Binning start time. If None, the start time is the first spike time.
    t_end : Optional[float]
        Binning end time. If None, the end time is the last spike time.

    Returns
    -------
        probability_distribution: np.ndarray
        probability distribution for the provided spiketrain

    """
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    prob_spike = (
        bin_spike.data.sum(axis=bin_spike._SIGNALAXIS, keepdims=True)
        / bin_spike.shape[bin_spike._SIGNALAXIS]
    )
    probability_distribution = np.empty_like(bin_spike.data, dtype=np.float64)
    _prob_spike = np.repeat(
        prob_spike, bin_spike.shape[bin_spike._SIGNALAXIS], axis=bin_spike._SIGNALAXIS
    )
    probability_distribution[bin_spike.data] = _prob_spike[bin_spike.data]
    probability_distribution[~bin_spike.data] = (1 - _prob_spike)[~bin_spike.data]
    return probability_distribution


@tag_info_metrics("self")
def shannon_entropy(
    spiketrains: Spikestamps,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the shannon entropy for a single channel recording using the binned spiketrain.

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
        shannon_entropy: float
        Shannon entropy for the given channel

    """
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    prob_spike = (
        bin_spike.data.sum(axis=bin_spike._SIGNALAXIS, keepdims=True)
        / bin_spike.shape[bin_spike._SIGNALAXIS]
    ).astype(np.float64)
    prob_no_spike = 1 - prob_spike
    shannon_entropy = -(
        prob_spike * np.log2(prob_spike) + prob_no_spike * np.log2(prob_no_spike)
    )
    return shannon_entropy


@tag_info_metrics("self")
def block_entropy(
    spiketrains: Spikestamps,
    history: int,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the block entropy for a recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    history : int
        history length
    bin_size : float
        bin size in seconds
    t_start : float
        Binning start time
    t_end : float
        Binning end time

    Returns
    -------
    block_entropy: float
        Block entropy for the given channel

    """
    import pyinform

    assert history > 0, "history length should be a finite positive value"
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    block_entropy = np.zeros((bin_spike.shape[bin_spike._CHANNELAXIS],))
    for idx, bin_spike_channel in enumerate(bin_spike):
        block_entropy[idx] = pyinform.blockentropy.block_entropy(
            bin_spike_channel, history
        )
    return block_entropy


@tag_info_metrics("self")
def entropy_rate(
    spiketrains: Spikestamps,
    history: float,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the entropy rate for a each channel recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    history : int
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    entropy rate: float
        entropy rate for the given channel

    """
    import pyinform

    assert history > 0, "history length should be a finite positive value"
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    entropy_rate = []
    for idx, bin_spike_channel in enumerate(bin_spike):
        entropy_rate.append(
            pyinform.entropyrate.entropy_rate(bin_spike_channel, k=history, local=True)[
                0
            ]
        )
    return np.asarray(entropy_rate), bin_spike  # TODO: return single type


@tag_info_metrics("self")
def active_information(
    spiketrains: Spikestamps,
    history: float,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the active information for a single channel recording using the binned spiketrain

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    bin_size : float
        bin size in seconds
    history : float
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time

    Returns
    -------
    active_information: float
        Active information for the given channel

    """
    import pyinform

    assert history > 0, "history length should be a finite positive value"
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)
    active_information = np.zeros((bin_spike.shape[bin_spike._CHANNELAXIS],))
    for idx, bin_spike_channel in enumerate(bin_spike):
        active_information[idx] = pyinform.activeinfo.active_info(
            bin_spike_channel, k=history
        )
    return active_information


@tag_info_metrics("pair")
def transfer_entropy(
    spiketrains: Spikestamps,
    history: float,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the transfer entropy for the pair of electorde recordings (X & Y) using the binned spiketrains and history

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    history : float
        history length
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    transfer_entropy: float
        Transfer_entropy for the given pair of electrodes

    """
    import pyinform

    assert history > 0, "history length should be a finite positive value"
    bin_spike = spiketrains.binning(bin_size=bin_size, t_start=t_start, t_end=t_end)

    transfer_entropy_matrix = np.zeros(
        [bin_spike.number_of_channels, bin_spike.number_of_channels]
    )
    for channelx in tqdm(range(bin_spike.number_of_channels)):
        for channely in tqdm(range(bin_spike.number_of_channels), leave=False):
            bin_spike_x = bin_spike[channelx]
            bin_spike_y = bin_spike[channely]

            # np.random.shuffle(bin_spike_x)

            normalizer = pyinform.entropyrate.entropy_rate(
                bin_spike_y, k=history, local=False
            )
            # normalizer = pyinform.mutualinfo.mutual_info(bin_spike_x, bin_spike_y, local=False)

            transfer_entropy = pyinform.transferentropy.transfer_entropy(
                bin_spike_x, bin_spike_y, k=history, local=False
            )
            if np.isclose(normalizer, 0):
                transfer_entropy_matrix[channelx, channely] = 0
            else:
                transfer_entropy_matrix[channelx, channely] = (
                    transfer_entropy / normalizer
                )
    return transfer_entropy_matrix


@tag_info_metrics("pair")
def mutual_information(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the mutual information for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    mutual_information: float
        Mutual information for the given pair of electrodes

    """
    import pyinform

    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"
    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    mutual_information = pyinform.mutualinfo.mutual_info(bin_spike_x, bin_spike_y)
    return mutual_information


@tag_info_metrics("pair")
def joint_entropy(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the joint entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    joint_entropy: float
        joint entropy for the given pair of electrodes

    """
    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    spike_dist_x = probability_distribution(
        spiketrains, channelx, t_start, t_end, bin_size
    )
    spike_dist_y = probability_distribution(
        spiketrains, channely, t_start, t_end, bin_size
    )
    spike_dist_xy = np.logical_and(spike_dist_x, spike_dist_y)
    joint_entropy = -np.sum(spike_dist_xy * np.log2(spike_dist_xy))
    return joint_entropy


@tag_info_metrics("pair")
def relative_entropy(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the relative entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    relative_entropy: float
        Relative_entropy for the given pair of electrodes

    """
    import pyinform

    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    relative_entropy = pyinform.relativeentropy.relative_entropy(
        bin_spike_x, bin_spike_y
    )
    return relative_entropy


@tag_info_metrics("pair")
def conditional_entropy(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the conditional entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    conditional_entropy: float
        conditional entropy for the given pair of electrodes

    """
    import pyinform

    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    bin_spike_x = binned_spiketrain(spiketrains, channelx, t_start, t_end, bin_size)
    bin_spike_y = binned_spiketrain(spiketrains, channely, t_start, t_end, bin_size)
    conditional_entropy = pyinform.conditionalentropy.conditional_entropy(
        bin_spike_x, bin_spike_y
    )
    return conditional_entropy


@tag_info_metrics("pair")
def cross_entropy(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Estimates the cross entropy for the pair of electorde recordings (X & Y) using the binned spiketrains

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    cross_entropy: float
        cross entropy for the given pair of electrodes

    """
    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    spike_dist_x = probability_distribution(
        spiketrains, channelx, t_start, t_end, bin_size
    )
    spike_dist_y = probability_distribution(
        spiketrains, channely, t_start, t_end, bin_size
    )
    cross_entropy = -np.sum(spike_dist_x * np.log2(spike_dist_y))
    return cross_entropy


@tag_info_metrics("pair")
def partial_information_decomposition(
    spiketrains: Spikestamps,
    channelx: float,
    channely: float,
    channelz: float,
    bin_size: float,
    t_start: float | None = None,
    t_end: float | None = None,
):
    """
    Decomposes the information provided by channel x and y about channel z in redundancy, unique information, and synergy.

    Parameters
    ----------
    spiketrains : Spikestamps
        Single spike-stamps
    channelx : float
        electrode/channel X
    channely : float
        electrode/channel Y
    channelz : float
        electrode/channel Y
    t_start : float
        Binning start time
    t_end : float
        Binning end time
    bin_size : float
        bin size in seconds

    Returns
    -------
    redundancy: float
        redundant information provided by both x and y about z
    unique_information_x: float
        information uniquely provided by x about z
    unique_information_y: float
        information uniquely provided by y about z
    synergy: float
        synergetic information provided by both x and y about z
    total_information: float
        total information provided by both x and y about z
    """
    assert t_start < t_end, "start time cannot be equal or greater than end time"
    assert bin_size > 0, "bin_size should be a finite positive value"

    spike_dist_x = probability_distribution(
        spiketrains, channelx, t_start, t_end, bin_size
    )
    spike_dist_y = probability_distribution(
        spiketrains, channely, t_start, t_end, bin_size
    )
    spike_dist_z = probability_distribution(
        spiketrains, channely, t_start, t_end, bin_size
    )
    spike_dist_xz = np.logical_and(spike_dist_x, spike_dist_z)
    spike_dist_yz = np.logical_and(spike_dist_y, spike_dist_z)
    spike_dist_xy = np.logical_and(spike_dist_x, spike_dist_y)
    spike_dist_xyz = np.logical_and(spike_dist_x, spike_dist_y, spike_dist_z)
    I_x_z = np.sum(
        spike_dist_xz * np.log2(spike_dist_xz / (spike_dist_x * spike_dist_z))
    )
    I_y_z = np.sum(
        spike_dist_yz * np.log2(spike_dist_yz / (spike_dist_y * spike_dist_z))
    )
    redundancy = np.min(I_x_z, I_y_z)
    unique_information_x = I_x_z - redundancy
    unique_information_y = I_y_z - redundancy
    total_information = np.sum(
        spike_dist_xyz * np.log2(spike_dist_xyz / (spike_dist_xy * spike_dist_z))
    )
    synergy = (
        total_information - redundancy - unique_information_x - unique_information_y
    )
    return (
        redundancy,
        unique_information_x,
        unique_information_y,
        synergy,
        total_information,
    )



__doc__ = """

Module (Intan)
##############

.. autoclass:: DataIntan
   :members:

.. autoclass:: DataIntanTriggered
   :members:

"""
__all__ = ["DataIntan", "DataIntanTriggered"]

from typing import Any, cast
from collections.abc import Iterable, Generator

import logging
import os
import xml.etree.ElementTree as ET
from glob import glob
from pathlib import Path

import numpy as np
from tqdm import tqdm

from miv.io.intan import rhs
from miv.core.datatype import Signal, Spikestamps
from miv.core.operator.wrapper import cached_method
from miv.io.openephys.data import Data


class DataIntan(Data):
    """Single data unit handler, collected from Intan.

    Each data unit that contains single recording. This class provides useful tools,
    such as masking channel, export data, interface with other packages, etc.
    If you have multiple recordings you would like to handle at the same time, use
    `DataManager` instead.

    By default recording setup, the following directory structure is expected in ``data_path``::

        data_<date>_<time>              # <- recorded data path
        ├── settings.xml
        ├── data_<date>_<time1>.rhs      # <- first rhs file
        ├── data_<date>_<time2>.rhs      # <- second rhs file
        ├── data_<date>_<time3>.rhs      # <- third rhs file
        ⋮
        └── analysis

        Parameters
        ----------
        data_path : str
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    def export(
        self,
        filename: str | Path,
        channels: Iterable[int] | None = None,
        progress_bar: bool = False,
    ) -> None:
        """
        Export data to the specified path.
        TODO: implement "view" time range

        Parameters
        ----------
        filename : str or Path
            Path to export data.
        """
        from miv.io import file as miv_file

        data = miv_file.initialize()
        miv_file.create_group(data, "Ephys", counter="nobj")
        miv_file.create_dataset(data, "Timestamps", group="Ephys", dtype=np.float32)
        miv_file.create_dataset(data, "Rate", group="Ephys", dtype=np.float32)
        miv_file.create_dataset(data, "Data", group="Ephys", dtype=np.float32)

        container = miv_file.create_container(data)
        data_shape = None
        for signal in tqdm(self.load(), disable=not progress_bar):
            if channels is None:
                matrix = signal.data
            else:
                channels = np.asarray(channels, dtype=np.int_)
                matrix = signal.data[:, channels]
            if data_shape is None:
                data_shape = matrix.shape
            elif data_shape != matrix.shape:
                break
            container["Ephys/Data"] = matrix
            container["Ephys/Timestamps"] = signal.timestamps
            container["Ephys/Rate"] = signal.rate
            test = miv_file.pack(data, container)
            assert test == 0

        miv_file.write(str(filename), data)

    def load(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Generator[Signal]:
        """
        Iterator to load data fragmentally.
        This function loads each file separately.

        Examples
        --------
            >>> data = Data(data_path)
            >>> for data.load(10) as (signal, timestamps, sampling_rate):
            ...     ...

        Returns
        -------
        signal : SignalType, neo.core.AnalogSignal
            The length of the first axis `signal.shape[0]` correspond to the length of the
            signal, while second axis `signal.shape[1]` correspond to the number of channels.
        timestamps : numpy array
        sampling_rate : float

        Raises
        ------
        FileNotFoundError
            If some key files are missing.
        """

        active_channels, total = self._get_active_channels()
        for sig in self._generator_by_channel_name("amplifier_data"):
            self._expand_channels(sig, active_channels, total)
            yield sig

    def get_stimulation(self, progress_bar: bool = False) -> Signal:
        """
        Load stimulation recorded data.
        """
        signals, timestamps = [], []
        sampling_rate: float
        for data in self._generator_by_channel_name("stim_data", progress_bar):
            signals.append(data.data)
            timestamps.append(data.timestamps)
            sampling_rate = data.rate

        signal = Signal(
            data=np.concatenate(signals, axis=0),
            timestamps=np.concatenate(timestamps),
            rate=sampling_rate,
        )
        active_channels, total = self._get_active_channels()
        self._expand_channels(signal, active_channels, total)
        return signal

    def _generator_by_channel_name(
        self, name: str, progress_bar: bool = False
    ) -> Generator[Signal]:
        if not self.check_path_validity():
            raise FileNotFoundError("Data directory does not have all necessary files.")
        files = self.get_recording_files()
        # Get sampling rate from setting file
        setting_path = os.path.join(self.data_path, "settings.xml")
        sampling_rate = int(ET.parse(setting_path).getroot().attrib["SampleRateHertz"])
        # Read each files
        for filename in tqdm(files, disable=not progress_bar):
            result, data_present = rhs.load_file(filename)
            assert data_present, f"Data does not present: {filename=}."
            assert not hasattr(result, name), f"No {name} in the file ({filename=})."

            signal = np.asarray(result[name]).T

            yield Signal(
                data=signal,
                timestamps=np.asarray(result["t"]),
                rate=sampling_rate,
            )

    def _get_active_channels(
        self, group_prefix: tuple[str, ...] = ("A", "B", "C", "D")
    ) -> tuple[np.ndarray, int]:
        setting_path = os.path.join(self.data_path, "settings.xml")
        root = ET.parse(setting_path).getroot()
        total = 0
        active_channels = []
        for child in root:
            if child.tag != "SignalGroup":
                continue
            if child.attrib["Prefix"] not in group_prefix:
                continue
            for channel in child:
                if channel.attrib["Enabled"] == "True":
                    active_channels.append(total)
                total += 1
        return np.array(active_channels), total

    def _expand_channels(
        self, signal: Signal, active_channels: np.ndarray, num_active_channels: int
    ) -> None:
        """
        Expand number of channels in `signal` to match the active channels.
        """
        if num_active_channels != signal.number_of_channels:
            _data = np.zeros(
                [signal.shape[0], num_active_channels], dtype=signal.data.dtype
            )
            if active_channels.size != signal.data.shape[1]:
                # TODO: figure out why this can happen
                self.logger.warning(
                    f"Size mismatch in the data: {active_channels.size} enabled channel vs {signal.data.shape=}"
                )
                if signal.data.shape[1] > active_channels.size:
                    _data[:, active_channels] = signal.data[:, active_channels.size]
                elif signal.data.shape[1] < active_channels.size:
                    _data[:, active_channels[: signal.data.shape[1]]] = signal.data
            else:
                _data[:, active_channels] = signal.data
            signal.data = _data

    def _read_header(self) -> dict[str, Any]:
        filename = self.get_recording_files()[0]
        fid = open(filename, "rb")
        header = rhs.read_header(fid)
        return header

    def check_path_validity(self) -> bool:
        """
        Check if necessary files exist in the directory.

        - Check `rhs` file exists.
        - Check `settings.xml` file exists.

        Returns
        -------
        bool
            Return true if all necessary files exist in the directory.
        """

        continuous_dat_paths = self.get_recording_files()
        if len(continuous_dat_paths) == 0:
            logging.warning("At least one .rhs file must exist in the data path.")
            return False
        if not os.path.exists(os.path.join(self.data_path, "settings.xml")):
            logging.warning("Missing settings.xml in the data path.")
            return False
        return True

    def get_recording_files(self) -> list[str]:
        """
        Get list of path of all recording files.
        """
        paths = glob(os.path.join(self.data_path, "*.rhs"), recursive=True)
        paths.sort()
        return paths

    def get_stimulation_events(self) -> Spikestamps | None:  # TODO: refactor
        """
        Get stimulation in Spikestamps form, where each stamps represent the stimulus event.
        """
        minimum_stimulation_length = 0.010
        data = self.get_stimulation()
        stim = data.data
        timestamps = data.timestamps
        # sampling_rate = data.rate
        stimulated_channels = np.where(np.abs(stim).sum(axis=0))[0]
        if len(stimulated_channels) == 0:
            return None
        stimulated_channel = stimulated_channels[0]
        stim = stim[:, stimulated_channel]

        events = ~np.isclose(stim, 0)
        eventstrain = timestamps[np.where(events)[0]]
        ref = np.concatenate(
            [[True], np.diff(eventstrain) > minimum_stimulation_length]
        )
        eventstrain = eventstrain[ref]
        ret = Spikestamps([eventstrain])  # TODO: use datatype.Events
        return ret

    def _load_digital_event_common(
        self, name: str, num_channels: int, progress_bar: bool = False
    ) -> Spikestamps:
        stamps: list[list[float]] = [[] for _ in range(num_channels)]
        for sig in self._generator_by_channel_name(name, progress_bar=progress_bar):
            for channel in range(num_channels):
                index = np.where(sig.data[:, channel])[0]
                stamps[channel].extend(np.asarray(sig.timestamps)[index])
        return Spikestamps(stamps)

    # @cached_method(cache_tag="digital_in")
    def load_digital_in_event(
        self,
        progress_bar: bool = False,
    ) -> Spikestamps:  # pragma: no cover
        """
        Load recorded data from digital input ports.
        Result is a list of timestamps for each channel, in Spikestamps format.

        Parameters
        ----------
        progress_bar : bool, optional
            Show progress bar, by default False
        """
        self.header = self._read_header()
        num_channels = self.header["num_board_dig_in_channels"]
        return self._load_digital_event_common(
            "board_dig_in_data", num_channels, progress_bar=progress_bar
        )

    # @cached_method(cache_tag="digital_out")
    def load_digital_out_event(
        self,
        progress_bar: bool = False,
    ) -> Spikestamps:  # pragma: no cover
        """
        Load recorded data from digital output ports.
        Result is a list of timestamps for each channel, in Spikestamps format.

        Parameters
        ----------
        progress_bar : bool, optional
            Show progress bar, by default False
        """
        self.header = self._read_header()
        num_channels = self.header["num_board_dig_out_channels"]
        return self._load_digital_event_common(
            "board_dig_out_data", num_channels, progress_bar=progress_bar
        )

    # @cached_method(cache_tag="ttl_events")
    def load_ttl_event(
        self,
        deadtime: float = 0.002,
        compress: bool = False,
        progress_bar: bool = False,
    ) -> Signal:
        """
        Load TTL events recorded data.

        Parameters
        ----------
        deadtime : float
            Deadtime between two TTL events. (default: 0.002)
        compress : bool
            If True, reduce rate of the signal. (default: False)
        progress_bar : bool
            If True, show progress bar. (default: False)
        """

        signals, timestamps = [], []
        sampling_rate: float
        active_channels, total = self._get_active_channels()
        for data in self._generator_by_channel_name("stim_data", progress_bar):
            self._expand_channels(data, active_channels, total)
            dead_time_idx = int(deadtime * data.rate)
            num_channels = data.number_of_channels

            signal = np.zeros((data.shape[data._SIGNALAXIS], 1), dtype=np.int_)
            for channel in range(num_channels):
                array = data.data[:, channel]
                threshold_crossings = np.nonzero(array)[0]
                if len(threshold_crossings) == 0:
                    continue

                distance_sufficient = np.insert(
                    np.diff(threshold_crossings) >= dead_time_idx, 0, True
                )
                while not np.all(distance_sufficient):
                    # repeatedly remove all threshold crossings that violate the dead_time
                    threshold_crossings = threshold_crossings[distance_sufficient]
                    distance_sufficient = np.insert(
                        np.diff(threshold_crossings) >= dead_time_idx, 0, True
                    )
                signal[threshold_crossings, 0] = channel
            signals.append(signal)
            timestamps.append(data.timestamps)
            sampling_rate = data.rate

        concatenated_arrays = np.concatenate(signals, axis=0)
        timestamps = np.concatenate(timestamps)

        if compress:  # TODO
            raise NotImplementedError

        return Signal(
            data=concatenated_arrays, timestamps=timestamps, rate=sampling_rate
        )


class DataIntanTriggered(DataIntan):
    """
    DataIntanTriggered is a subclass of DataIntan, which is used to handle Intan
    recording when the recording is triggered by TTL signals.
    """

    def __init__(
        self,
        data_path: str,  # FIXME: argument order with intan.DATA
        index: int = 0,
        trigger_key: str = "board_adc_data",
        trigger_index: int = 0,
        trigger_threshold_voltage: float = 1.0,
        progress_bar: bool = False,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, data_path=data_path, **kwargs)
        self.index = index
        self.trigger_key = trigger_key
        self.trigger_index = trigger_index
        self.trigger_threshold_voltage = trigger_threshold_voltage
        self.progress_bar = progress_bar

    def __len__(self) -> int:
        groups = self._trigger_grouping()
        return len(groups)

    def __getitem__(self, index: int) -> "DataIntanTriggered":
        groups = self._trigger_grouping()
        if len(groups) <= index:
            raise IndexError(
                f"Index exceeds the number of triggered recordings ({len(groups)})."
            )
        return DataIntanTriggered(
            data_path=self.data_path,
            index=index,
            trigger_key=self.trigger_key,
            trigger_index=self.trigger_index,
            trigger_threshold_voltage=self.trigger_threshold_voltage,
            progress_bar=self.progress_bar,
        )

    @cached_method(cache_tag="trigger_grouping")
    def _trigger_grouping(self) -> list[dict[str, list]]:
        def _find_sequence(arr: np.ndarray) -> np.ndarray:
            if arr.size == 0:
                return arr
            selected = arr[np.concatenate([np.array([True]), arr[1:] - 1 != arr[:-1]])]
            return cast(np.ndarray, selected)

        paths = DataIntan.get_recording_files(self)

        group_files = []
        group: dict[str, list] = {"paths": [], "start index": [], "end index": []}
        status = 0
        for file in tqdm(paths, disable=not self.progress_bar):
            result, _ = rhs.load_file(file)
            time = result["t"]

            self.logger.info(f"Recorded result keys: {list(result.keys())}")
            self.logger.info(f"Number of trigger: {len(result[self.trigger_key])}")

            # TODO: Something can be done better
            if "adc" in self.trigger_key:
                adc_data = result[self.trigger_key][self.trigger_index]
                diff_adc_data = adc_data[1:] - adc_data[:-1]
                recording_on = _find_sequence(
                    np.where(diff_adc_data > self.trigger_threshold_voltage)[0]
                ).tolist()
                recording_off = _find_sequence(
                    np.where(diff_adc_data < -self.trigger_threshold_voltage)[0]
                ).tolist()
            elif "dig" in self.trigger_key:
                recording_state = result[self.trigger_key][self.trigger_index]
                # recording_state = np.logical_xor(digital_data[1:], digital_data[:-1])
                recording_on = _find_sequence(
                    np.where(
                        np.logical_and(recording_state[1:], ~recording_state[:-1])
                    )[0]
                ).tolist()
                recording_off = _find_sequence(
                    np.where(
                        np.logical_and(~recording_state[1:], recording_state[:-1])
                    )[0]
                ).tolist()

            sindex = 0
            if status == 1 and len(recording_on) == 0 and len(recording_off) == 0:
                group["paths"].append(file)
                group["start index"].append(sindex)
                group["end index"].append(time.shape[0])
            while len(recording_on) > 0 or len(recording_off) > 0:
                if status == 0 and len(recording_on) > 0:
                    sindex = recording_on.pop(0)
                    status = 1
                    if len(recording_off) == 0:
                        group["paths"].append(file)
                        group["start index"].append(sindex)
                        group["end index"].append(time.shape[0])
                elif status == 1 and len(recording_off) > 0:
                    eindex = recording_off.pop(0)
                    status = 0
                    if eindex < sindex:
                        raise ValueError(
                            f"Something went wrong with the trigger signal. Starting index {sindex} must be less than ending index {eindex}."
                        )
                    group["paths"].append(file)
                    group["start index"].append(sindex)
                    group["end index"].append(eindex)
                    group_files.append(group)
                    group = {"paths": [], "start index": [], "end index": []}
                else:
                    raise ValueError(
                        f"Something went wrong with the trigger signal. {len(recording_on)=} {len(recording_off)=}"
                    )
        return group_files

    def get_recording_files(self) -> list[str]:
        groups = self._trigger_grouping()
        return groups[self.index]["paths"]

    def _generator_by_channel_name(
        self, name: str, progress_bar: bool = False
    ) -> Generator[Signal]:
        # TODO: move out _get_active_channels
        if not self.check_path_validity():
            raise FileNotFoundError("Data directory does not have all necessary files.")
        groups = self._trigger_grouping()
        files = groups[self.index]["paths"]
        sindex = groups[self.index]["start index"]
        eindex = groups[self.index]["end index"]
        # Get sampling rate from setting file
        setting_path = os.path.join(self.data_path, "settings.xml")
        sampling_rate = int(ET.parse(setting_path).getroot().attrib["SampleRateHertz"])
        active_channels, total = self._get_active_channels()
        # Read each files
        for filename, sidx, eidx in tqdm(
            zip(files, sindex, eindex, strict=False), disable=not progress_bar
        ):
            result, data_present = rhs.load_file(filename)
            assert data_present, f"Data does not present: {filename=}."
            assert not hasattr(result, name), f"No {name} in the file ({filename=})."

            signal = np.asarray(result[name])
            if total != signal.shape[0]:
                _signal = np.zeros([total, signal.shape[1]], dtype=signal.dtype)
                _signal[active_channels, :] = signal
                signal = _signal

            yield Signal(
                data=signal.T[sidx:eidx, :],
                timestamps=np.asarray(result["t"])[sidx:eidx],
                rate=sampling_rate,
            )



__doc__ = ""
__all__ = ["ButterBandpass"]

import os
import time
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as sps

from miv.core.datatype import Signal
from miv.core.operator_generator.operator import GeneratorOperatorMixin
from miv.core.operator_generator.wrapper import cache_generator_call


@dataclass
class ButterBandpass(GeneratorOperatorMixin):
    """Classical bandpass filter using `scipy` butterworth filter

    Parameters
    ----------
    lowcut : float
        low-pass frequency
    highcut : float
        high-pass frequency
    order : int
        The order of the filter. (default=5)
    tag : str
        Tag for the collection of filter.
    btype : str
        Filter type: bandpass, lowpass, highpass, bandstop. (default="bandpass")
        If set to lowpass, the critical frequency is set to 'highcut'.
        If set to highpass, the critical frequency is set to 'lowcut'.
        If set to bandpass or bandstop, the critical frequency is '[lowcut, highcut]'.
    """

    lowcut: float | None = None
    highcut: float | None = None
    order: int = 5
    tag: str = "bandpass filter"
    btype: str = "bandpass"

    @cache_generator_call
    def __call__(self, signal: Signal) -> Signal:
        """__call__.

        Parameters
        ----------
        signal : SignalType
            signal

        Returns
        -------
        Signal

        """
        rate = signal.rate
        sos = self._butter_bandpass(rate)
        y = signal.data.copy()
        num_channel = signal.number_of_channels

        durations = []
        for ch in range(num_channel):
            stime = time.time()
            y[:, ch] = sps.sosfiltfilt(sos, signal.data[:, ch])

            # logging
            duration = time.time() - stime
            durations.append(duration)
            # self.logger.info(f"Filtering channel {ch} took {duration:.3f} sec.")
        # log statistics - average, min, max duration
        self.logger.info(
            f"Filtering took {np.mean(durations):.3f} sec on average. "
            f"Min: {np.min(durations):.3f} sec, "
            f"Max: {np.max(durations):.3f} sec."
        )
        return Signal(data=y, timestamps=signal.timestamps, rate=rate)

    def __post_init__(self):
        if self.lowcut is not None and self.highcut is not None:
            assert self.lowcut < self.highcut, (
                f"{self.lowcut=} should be lower than {self.highcut=}."
            )
            assert min(self.lowcut, self.highcut) > 0, (
                "Filter critical frequencies must be greater than 0"
            )
        elif self.lowcut is not None:
            assert self.lowcut > 0, "Filter frequencies must be greater than 0"
        elif self.highcut is not None:
            assert self.highcut > 0, "Filter frequencies must be greater than 0"
        assert self.order > 0 and isinstance(self.order, int), (
            f"Filter {self.order} must be an nonnegative integer."
        )
        assert self.lowcut is not None or self.highcut is not None, (
            "Filtering frequency cannot be both None"
        )
        super().__init__()
        self.cacher.policy = "OFF"

    def _butter_bandpass(self, sampling_rate: float):
        if self.btype in ["bandpass", "bandstop"]:
            low = self.lowcut
            high = self.highcut
            critical_frequency = [low, high]
        elif self.btype == "highpass":
            critical_frequency = self.lowcut
        elif self.btype == "lowpass":
            critical_frequency = self.highcut
        else:
            raise ValueError("Unknown btype: %s" % self.btype)
        sos = sps.butter(
            self.order,
            critical_frequency,
            fs=sampling_rate,
            btype=self.btype,
            output="sos",
        )
        return sos

    def firstiter_plot_frequency_response(
        self,
        signal,
        inputs,
        show=False,
        save_path=None,
    ):
        """plot_frequency_response"""
        rate = signal.rate
        sos = self._butter_bandpass(rate)
        w, h = sps.sosfreqz(sos, worN=2000, fs=rate)
        ah = np.abs(h)

        fig = plt.figure()
        plt.semilogx(w, 20 * np.log10(np.where(ah > 0.0, ah, 1e-300)))
        plt.title("Butterworth filter frequency response")
        plt.xlabel("Frequency [Hz]")
        plt.ylabel("Amplitude [dB]")
        plt.margins(0, 0.1)
        plt.grid(which="both", axis="both")
        if self.lowcut is not None:
            plt.axvline(self.lowcut, color="red")
        if self.highcut is not None:
            plt.axvline(self.highcut, color="red")
        plt.ylim(-50, 10)

        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "filter_frequency_response.png"))

        plt.close(fig)



# A script to convert Open Ephys data to MiV HDF5 format.

import logging
import os
import sys

import click

from miv.io import DataManager
from miv.io import file as miv_file


def config_logging(verbose):
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARN)


script_name = os.path.basename(__file__)


def seq_find(f, seq):
    """
    Determines if an element satisfying predicate is present in the
    given sequence, returns index of element or None.

    :param f: predicate function
    :param seq: sequence
    :return: index of element satisfying predicate, or None

    """
    i = 0
    for x in seq:
        if f(x):
            return i
        else:
            i = i + 1
    return None


@click.command()
@click.option("--folder-path", "-p", required=True, type=click.Path())
@click.option("--verbose", "-v", type=bool, default=False, is_flag=True)
def main(folder_path: str, verbose: bool):
    """Convert Open Ephys data to MiV HDF5 format"""

    config_logging(verbose)
    logger = logging.getLogger("miv_file")

    # Load dataset from OpenEphys recording
    data_manager = DataManager(folder_path)
    miv_data = miv_file.initialize()
    cont = miv_file.create_container(miv_data)

    # Get signal and rate(hz)
    #   signal        : np.array, shape(N, N_channels)
    #   timestamps    : np.array
    #   sampling_rate : float
    for recording in data_manager:
        data_path = recording.data_path
        with recording.load() as (signal, timestamps, sampling_rate):
            group_id = miv_file.create_group(miv_data, data_path, counter="nrecording")
            miv_file.create_dataset(
                miv_data,
                ["signal", "timestamps", "sampling_rate"],
                group=data_path,
                dtype=float,
            )
            cont[f"{group_id}/signal"] = signal
            cont[f"{group_id}/timestamps"] = timestamps
            cont[f"{group_id}/sampling_rate"] = [sampling_rate]
            miv_file.pack(miv_data, cont, logger=logger)

    miv_file.write(
        f"{folder_path}/MiV_data.h5",
        miv_data,
        comp_type="gzip",
        comp_opts=9,
        logger=logger,
    )


if __name__ == "__main__":
    main(
        args=sys.argv[
            (seq_find(lambda x: os.path.basename(x) == script_name, sys.argv) + 1) :
        ]
    )



from __future__ import annotations

__doc__ = """"""
__all__ = ["BaseChainingMixin"]

from typing import TYPE_CHECKING, Any
from collections.abc import Iterator

import itertools

import networkx as nx
import matplotlib.pyplot as plt


if TYPE_CHECKING:
    from .protocol import _Chainable


class BaseChainingMixin:
    """
    Base mixin to create chaining structure between objects.

    Need further implementation of: output, tag
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._downstream_list: list[_Chainable] = []
        self._upstream_list: list[_Chainable] = []

    def __rshift__(self, right: _Chainable) -> _Chainable:
        self.append_downstream(right)
        right.append_upstream(self)
        return right

    def clear_connections(self) -> None:
        """Clear all the connections to other nodes, and remove dependencies."""
        for node in self.iterate_downstream():
            node.disconnect_upstream(self)
        for node in self.iterate_upstream():
            node.disconnect_downstream(self)
        self._downstream_list.clear()
        self._upstream_list.clear()

    def disconnect_upstream(self, node: _Chainable) -> None:
        self._upstream_list.remove(node)

    def disconnect_downstream(self, node: _Chainable) -> None:
        self._downstream_list.remove(node)

    def append_upstream(self, node: _Chainable) -> None:
        self._upstream_list.append(node)

    def append_downstream(self, node: _Chainable) -> None:
        self._downstream_list.append(node)

    def iterate_downstream(self) -> Iterator[_Chainable]:
        return iter(self._downstream_list)

    def iterate_upstream(self) -> Iterator[_Chainable]:
        return iter(self._upstream_list)

    def visualize(self, show: bool = False, seed: int = 200) -> nx.DiGraph:
        """
        Visualize the network structure of the "Operator".
        """
        G = nx.DiGraph()

        # BFS
        visited: list[_Chainable] = []
        next_list: list[_Chainable] = [self]
        while next_list:
            v = next_list.pop()
            visited.append(v)
            for node in itertools.chain(v.iterate_downstream()):
                G.add_edge(repr(v), repr(node))
                if node in visited or node in next_list:
                    continue
                visited.append(node)
                next_list.append(node)

        # Draw the graph
        # TODO: Balance layout
        # TODO: On edge, label the argument order
        pos = nx.spring_layout(G, seed=seed)
        nx.draw_networkx_nodes(G, pos, node_size=500, alpha=0.8)
        nx.draw_networkx_edges(G, pos, width=2, arrows=True)
        nx.draw_networkx_labels(G, pos, font_size=12, font_family="sans-serif")

        # Display the graph
        plt.margins(x=0.4)
        plt.axis("off")
        if show:
            plt.show()
        return G

    def _text_visualize_hierarchy(
        self,
        string_list: list[tuple[int, _Chainable]],
        prefix: str = "|__ ",
    ) -> str:
        output = []
        for _i, item in enumerate(string_list):
            depth, label = item
            if depth > 0:
                output.append("    " * (depth - 1) + prefix + str(label))
            else:
                output.append(str(label))
        return "\n".join(output)



__all__ = ["GridMEA"]


import MEAutility
import numpy as np

from miv.core.datatype import Signal
from miv.mea.base import MEAMixin


class GridMEA(MEAMixin, MEAutility.core.MEA):
    """
    A class representing a grid-based multi-electrode array (MEA).

    Example
    -------
    The best way to create MEA module is through `miv.mea.MEA` object.
    For example, you could create an instance of the GridMEA class like this:
    literal blocks::

        mea: GridMEA = miv.mea.MEA.return_mea("64_intanRHD")
    """

    def __init__(
        self,
        positions,  # Exclude all "None" electrodes
        *args,
        **kwargs,
    ):
        positions = positions.astype(np.float64)
        info = kwargs["info"]

        # Remove nan channels - in json, nan channels are open or broken channels
        MEAMixin.__init__(
            self, positions=positions[~np.isnan(positions).any(axis=1)], *args, **kwargs
        )

        self.grid = np.zeros(info["dim"], dtype=int) - 1
        for channel in range(positions.shape[0]):
            x, y, _ = positions[channel]
            if np.isnan(x) or np.isnan(y):
                continue
            ix = int(np.round(x / self.pitch[0]))
            iy = int(np.round(y / self.pitch[1]))
            self.grid[iy, ix] = channel
        self.grid = self.grid[::-1, :]

        self.nrow, self.ncol = self.grid.shape
        self.X = np.linspace(0, self.pitch[0] * self.ncol, self.ncol)
        self.Y = np.linspace(0, -self.pitch[1] * self.nrow, self.nrow)
        self.Xn, self.Yn = np.meshgrid(self.X, self.Y)

    @property
    def channels(self):
        return sorted(self.grid[self.grid >= 0].ravel().tolist())

    def to_json(self) -> dict:
        """Return a JSON-serializable dictionary"""
        return {
            "grid": self.grid.tolist(),
            "pitch": self.pitch,
        }

    def map_data(
        self, vector: np.ndarray, missing_value: float = 0.0
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Map data (1-D array) to MEA (2-D array or N-D array)"""
        value_grid = np.full_like(self.grid, missing_value, dtype=np.float64)
        for idx, value in enumerate(vector):
            if idx not in self.grid:
                continue
            value_grid[self.grid == idx] = value

        return self.Xn, self.Yn, value_grid

    def map_temporal_data(self, signal: Signal, missing_value: float = 0.0):
        """Map signal data to MEA"""
        n_time = signal.shape[signal._SIGNALAXIS]
        value_grid = np.full(
            [n_time, self.nrow, self.ncol], missing_value, dtype=np.float64
        )
        for idx in range(signal.number_of_channels):
            if idx not in self.grid:
                continue
            value_grid[:, self.grid == idx] = signal[idx][:, None]
        return self.Xn, self.Yn, value_grid

    def get_ixiy(self, channel: int):
        # FIXME: ixiy return ys xs
        """Given node index, return x y coordinate index"""
        if channel not in self.grid:
            return None
        ys, xs = np.where(self.grid == channel)
        assert len(xs) == 1 and len(ys) == 1, f"The index {channel} is not unique."
        return ys[0], xs[0]

    def get_xy(self, channel: int):
        """Given node index, return x y coordinate index"""
        if channel not in self.grid:
            return None
        ys, xs = np.where(self.grid == channel)
        assert len(xs) == 1 and len(ys) == 1, f"The index {channel} is not unique."
        return self.Y[ys[0]], self.X[xs[0]]

    @property
    def coordinates(self):
        """Return the coordinates of the electrodes

        https://numpy.org/doc/stable/reference/generated/numpy.meshgrid.html
        """

        return np.vstack([self.Xn.ravel(), self.Yn.ravel()]).T

    def figsize(self, scale=(4, 4)):
        """Return an ideal figure size for plt.figure"""
        return (self.ncol * scale[0], self.nrow * scale[1])

    # Override: get_electrode_matrix
    def get_electrode_matrix(self):
        return self.grid



__all__ = ["MEA"]


import logging
import os

import MEAutility
import yaml

from miv.mea.grid import GridMEA

_MEA_MODULE: type[MEAutility.core.MEA] = MEAutility.core.MEA
_RECTMEA_MODULE: type[MEAutility.core.MEA] = GridMEA  # MEAutility.core.RectMEA


class _MEA:
    """
    Motivated module as MEAutility.core
    """

    _reg_electrode_paths = [
        os.path.join(os.path.split(MEAutility.__file__)[0], "electrodes"),  # default
        os.path.join(os.path.split(__file__)[0], "electrodes"),  # for development
    ]

    # Make it singleton class
    _instance = None

    def __init__(self):
        """
        Initialize MEA object.
        """
        self.logger = logging.getLogger(__name__)

    def __new__(cls, *args, **kwargs):
        """
        Make it singleton class.
        """
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __call__(self):
        """
        Show all available MEA models.
        """
        self.return_mea_list(verbose=True)

    def register_electrode_path(self, path):
        """
        Register path for electrode definition files.

        Parameters
        ----------
        path: str
            Path to folder with electrode definition files.
        """
        self._reg_electrode_paths.append(path)

    def get_electrode_paths(self):
        """
        Get paths for electrode definition files in yaml.
        """
        paths = []
        for folder in self._reg_electrode_paths:
            if not os.path.exists(folder):
                continue
            for f in os.listdir(folder):
                # check stem
                if not f.endswith(".yml") and not f.endswith(".yaml"):
                    continue
                paths.append(os.path.join(folder, f))
        return paths

    def return_mea(self, electrode_name: str):
        """
        Returns MEA object.

        Parameters
        ----------
        electrode_name: str
            Probe/MEA name

        Returns
        -------
        mea: MEA
            The MEA object
        """
        mea = None
        counter = 0

        # load MEA info
        for elinfo in self.iter_electrode_info():
            if elinfo["electrode_name"] == electrode_name:
                if counter > 1:
                    self.logger.warning(
                        f"Found multiple MEA models named {electrode_name}."
                        " Returning the last one."
                    )
                pos = MEAutility.core.get_positions(elinfo, False)
                # create MEA object
                if MEAutility.core.check_if_rect(elinfo):
                    mea = _RECTMEA_MODULE(positions=pos, info=elinfo)
                else:
                    mea = _MEA_MODULE(positions=pos, info=elinfo)
                counter += 1

        if mea is None:
            self.logger.error(
                f"MEA model named {electrode_name} not found\n"
                f"Available models: {self.__call__()}"
            )
        return mea

    def return_mea_from_dict(self, info: dict):
        """
        Build MEA object from info dictionary.

        Parameters
        ----------
        info: dict
            Dictionary with electrode info

        Returns
        -------
        mea
        """
        elinfo = info
        if "center" in elinfo.keys():
            center = elinfo["center"]
        else:
            center = True
        pos = MEAutility.core.get_positions(elinfo, center=center)
        # create MEA object
        if MEAutility.core.check_if_rect(elinfo):
            mea = _RECTMEA_MODULE(positions=pos, info=elinfo)
        else:
            mea = _MEA_MODULE(positions=pos, info=elinfo)

        return mea

    def return_mea_info(self, electrode_name=None):
        """
        Returns probe information.

        Parameters
        ----------
        electrode_name: str
            Probe name

        Returns
        -------
        info: dict
            Dictionary with electrode info
        """
        # load MEA info
        for elinfo in self.iter_electrode_info():
            if elinfo.get("electrode_name") == electrode_name:
                return elinfo

    def return_mea_list(self, verbose=False):
        """
        Returns available probe models.

        Returns
        -------
        probes: List[str]
            List of available probe_names
        """
        electrode_names = [
            info.get("electrode_name", "") for info in self.iter_electrode_info()
        ]

        if verbose:
            self.logger.info("Available MEA models:")
            for probe in electrode_names:
                self.logger.info(probe)

        return sorted(electrode_names)

    def iter_electrode_info(self):
        """
        Iterate all electrode info.
        """
        for fname in self.get_electrode_paths():
            with open(fname) as meafile:
                if MEAutility.core.use_loader:
                    elinfo = yaml.load(meafile, Loader=yaml.FullLoader)
                else:
                    elinfo = yaml.load(meafile)
            if elinfo is None:
                # Warn incompatible yaml file
                self.logger.warning(f"File {fname} is not compatible with yaml.load")
            else:
                yield elinfo


MEA = _MEA()



__all__ = [
    "VanillaGeneratorRunner",
]

from typing import Any
import inspect
from itertools import islice
import time
from collections.abc import Generator

import multiprocessing as mp

from ..protocol import _LazyCallable


class VanillaGeneratorRunner:
    """Default runner without any modification.
    Simply, the operator will be executed in embarrassingly parallel manner.
    If MPI is available, the operator will be executed in every ranks.

    This runner is meant to be used for generator operators.
    """

    def __init__(self, parent):
        self.parent = parent

    def __call__(
        self, func: _LazyCallable, inputs: list[Generator[Any]] | None = None
    ) -> Generator:
        is_all_generator = all(inspect.isgenerator(v) for v in inputs)
        if not is_all_generator:
            result = func(self, *inputs)  # FIXME: this will probably cause issue
            return result

        def generator_func(*args: tuple[Generator, ...]) -> Generator:
            num_workers = 1
            tasks = zip(*args, strict=False)
            for idx, zip_arg in enumerate(tasks):
                stime = time.time()
                result = func(idx, *zip_arg)
                print(f"    iter {idx:03d} {self}: {time.time() - stime:.03f}sec", flush=True)
                yield result
            print("generator-tasks done", flush=True)

            # TODO: add lastiter_plot
            # FIXME
            self.parent._done_flag_generator_plot = True
            self.parent._done_flag_firstiter_plot = True

        generator = generator_func(*inputs)
        return generator

        # return func(*inputs)  # type: ignore

    def get_run_order(self) -> int:
        return 0

class GeneratorRunnerInMultiprocessing:
    """Default runner without any modification.
    Simply, the operator will be executed in embarrassingly parallel manner.
    If MPI is available, the operator will be executed in every ranks.

    This runner is meant to be used for generator operators.
    """

    def __init__(self, parent, chunk_size:int=4):
        self.parent = parent
        self.chunk_size = chunk_size

    def __call__(
        self, func: _LazyCallable, inputs: list[Generator[Any]] | None = None
    ) -> Generator:
        is_all_generator = all(inspect.isgenerator(v) for v in inputs)
        if not is_all_generator:
            result = func(self, *inputs)  # FIXME: this will probably cause issue
            return result

        def generator_func(*args: tuple[Generator, ...]) -> Generator:
            num_workers = self.chunk_size
            istart = 0
            tasks = zip(*args, strict=False)
            with mp.Pool(processes=num_workers) as pool:
                while zip_arg := list(islice(tasks, num_workers)):
                    proxy_func = getattr(self.parent.__class__, func.__name__)
                    stime = time.time()
                    _args = [
                        tuple([self.parent] + [istart + i] + list(za))
                        for i, za in enumerate(zip_arg)
                    ]
                    # results = pool.imap(prox_func, _args)
                    results = pool.starmap(proxy_func, _args)
                    istart += len(_args)
                    print(
                        f"completed tasks: {istart}(+{len(_args)}) ({time.time() - stime:.2f}sec)",
                        flush=True,
                    )

                    stime = time.time()
                    yield from results
                    print(
                        f"external_tasks:  ({time.time() - stime:.2f}sec)", flush=True
                    )

            print("generator-tasks done", flush=True)

            # for idx, zip_arg in enumerate(zip(*args, strict=False)):
            #    stime = time.time()
            #    result = func(self, *zip_arg, idx=idx)
            # print(f"    iter {idx:03d} {self}: {time.time() - stime:.03f}sec", flush=True)
            #    yield result
            # TODO: add lastiter_plot
            # FIXME
            self.parent._done_flag_generator_plot = True
            self.parent._done_flag_firstiter_plot = True

        generator = generator_func(*inputs)
        return generator

        # return func(*inputs)  # type: ignore

    def get_run_order(self) -> int:
        return 0



from __future__ import annotations

__all__ = [
    "_RunnerProtocol",
    "VanillaRunner",
    "SupportMultiprocessing",
    "MultiprocessingRunner",
    "InternallyMultiprocessing",
    "StrictMPIRunner",
    "SupportMPIMerge",
]
from typing import TYPE_CHECKING, Any, Protocol, cast
import multiprocessing
from collections.abc import Callable, Generator

if TYPE_CHECKING:
    # This will likely cause circular import error
    from miv.core.datatype import DataTypes
    import mpi4py


class _RunnerProtocol(Protocol):
    def __init__(
        self,
        *,
        comm: mpi4py.MPI.Comm | None = None,
        root: int = 0,
        **kwargs: Any,
    ) -> None: ...

    def __call__(
        self, func: Callable, inputs: Any | None = None
    ) -> Generator[Any] | Any: ...

    def get_run_order(self) -> int:
        """
        The method determines the order of execution, useful for
        multiprocessing or MPI.
        """
        ...


class VanillaRunner:
    """Default runner without any high-level parallelism.
    Simply, the operator will be executed in root-rank, and distributed across other ranks.
    If MPI is not available, the operator will be executed in root-rank only.
    """

    def __init__(self, *, comm: mpi4py.MPI.Comm | None = None, root: int = 0) -> None:
        self.comm: mpi4py.MPI.Comm | None
        self.is_root: bool

        if comm is not None:
            self.comm = comm
            self.is_root = self.comm.Get_rank() == root
        else:
            try:
                from mpi4py import MPI

                self.comm = MPI.COMM_WORLD
                self.is_root = self.comm.Get_rank() == 0
                if self.comm.Get_size() == 1:
                    self.comm = None
                    self.is_root = True
            except ImportError:
                self.comm = None
                self.is_root = True

    def get_run_order(self) -> int:
        return 0

    def _execute(self, func: Callable, inputs: DataTypes) -> DataTypes:
        if inputs is None:
            output = func()
        else:
            inputs = cast("DataTypes", inputs)
            output = func(*inputs)
        return output

    def __call__(self, func: Callable, inputs: DataTypes | None = None) -> DataTypes:
        output = None
        if self.is_root:
            output = self._execute(func, inputs)

        if self.comm is not None:
            output = self.comm.bcast(output, root=0)
        return output


class MultiprocessingRunner:
    def __init__(self, *, np: int | None = None) -> None:
        if np is None:
            self._np = multiprocessing.cpu_count()
        else:
            self._np = np

    def get_run_order(self) -> int:
        return 0  # FIXME

    @property
    def num_proc(self) -> int:
        return self._np

    def __call__(
        self, func: Callable, inputs: Generator[DataTypes] | None = None
    ) -> Generator[DataTypes]:
        if inputs is None:
            raise NotImplementedError(
                "Multiprocessing for operator with no generator input is not supported yet. Please use VanillaRunner for this operator."
            )
        else:
            with multiprocessing.Pool(self.num_proc) as p:
                yield from p.imap(func, inputs)


class StrictMPIRunner:
    def __init__(self, *, comm: mpi4py.MPI.Comm, root: int = 0) -> None:
        self.comm = comm
        self.root = root

    def get_run_order(self) -> int:
        return int(self.get_rank())

    def get_rank(self) -> int:
        return int(self.comm.Get_rank())

    def get_size(self) -> int:
        return int(self.comm.Get_size())

    def get_root(self) -> int:
        return self.root

    def is_root(self) -> bool:
        return self.get_rank() == self.root

    def __call__(self, func: Callable, inputs: DataTypes | None = None) -> DataTypes:
        if inputs is None:
            output = func()
        else:
            inputs = cast(tuple["DataTypes"], inputs)
            output = func(*inputs)
        return output


class SupportMPIMerge(StrictMPIRunner):
    """
    This runner policy is used for operators that can be merged by MPI.
    """

    def __call__(self, func: Callable, inputs: DataTypes | None = None) -> DataTypes:
        if inputs is None:
            output = func()
        else:
            inputs = cast(tuple["DataTypes"], inputs)
            output = func(*inputs)

        outputs = self.comm.gather(output, root=self.root)
        if self.is_root():
            result = output.from_collapse(outputs)  # Class method
        else:
            result = None
        result = self.comm.bcast(result, root=self.root)
        return result


class SupportMPIWithoutBroadcast(StrictMPIRunner):
    """
    This runner policy is used for operators that can be merged by MPI.
    """

    def __call__(
        self, func: Callable, inputs: tuple[DataTypes] | None = None
    ) -> DataTypes | None:
        if inputs is None:
            output = func()
        else:
            output = func(*inputs)

        outputs = self.comm.gather(output, root=self.root)
        if self.is_root():
            result = output.from_collapse(outputs)  # Class method
        else:
            result = None
        return result


class SupportMultiprocessing:
    pass


class InternallyMultiprocessing(Protocol):
    @property
    def num_proc(self) -> int: ...



__doc__ = ""
__all__ = ["Notch"]


import os
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import scipy.signal as sps

from miv.core.datatype import Signal
from miv.core.operator_generator.operator import GeneratorOperatorMixin
from miv.core.operator_generator.wrapper import cache_generator_call


@dataclass
class Notch(GeneratorOperatorMixin):
    """Notch filter using `scipy` second-order notch filter, wrapped in operator mixin.

    Parameters
    ----------
    w0 : float
        notch frequency
    Q : float
        Q factor
    tag : str
        Tag for the collection of filter.
    """

    f0: float | None = 60.0  # Hz
    Q: float | None = 30.0
    tag: str = "notch filter"

    @cache_generator_call
    def __call__(self, signal: Signal) -> Signal:
        """__call__.

        Parameters
        ----------
        signal : SignalType
            signal

        Returns
        -------
        Signal

        """
        rate = signal.rate
        b, a = sps.iirnotch(w0=self.f0, Q=self.Q, fs=rate)
        y = signal.data.copy()
        num_channel = signal.number_of_channels
        for ch in range(num_channel):
            y[:, ch], h = sps.freqz(b, a, signal.data[:, ch])
        return Signal(data=y, timestamps=signal.timestamps, rate=rate)

    def __post_init__(self):
        super().__init__()
        self.cacher.policy = "OFF"

    def plot_frequency_response(self, signal, inputs, show=False, save_path=None):
        """plot_frequency_response"""
        rate = next(signal).rate
        b, a = sps.iirnotch(w0=self.f0, Q=self.Q, fs=rate)
        freq, h = sps.freqz(b, a, fs=rate)

        fig, ax = plt.subplots(2, 1, figsize=(8, 6))

        ax[0].plot(freq, 20 * np.log10(abs(h)))
        ax[0].set_title("Frequency Response")
        ax[0].set_ylabel("Amplitude (dB)")
        ax[0].set_xlim([0, 100])
        ax[0].set_ylim([-25, 10])
        ax[0].grid(True)

        ax[1].plot(freq, np.unwrap(np.angle(h)) * 180 / np.pi)
        ax[1].set_ylabel("Angle (degrees)")
        ax[1].set_xlabel("Frequency (Hz)")
        ax[1].set_xlim([0, 100])
        ax[1].set_yticks([-90, -60, -30, 0, 30, 60, 90])
        ax[1].set_ylim([-90, 90])
        ax[1].grid(True)

        if show:
            plt.show()
        if save_path is not None:
            plt.savefig(os.path.join(save_path, "filter_frequency_response.png"))

        plt.close(fig)



import importlib
import pkgutil

_discovered_plugins = [
    name for finder, name, ispkg in pkgutil.iter_modules() if name.startswith("miv_")
]

for _plugin in _discovered_plugins:
    mdl = importlib.import_module(_plugin)

    if "__all__" in mdl.__dict__:
        names = mdl.__dict__["__all__"]
    else:
        names = [x for x in mdl.__dict__ if not x.startswith("_")]

    # now drag them in
    loads = {k: getattr(mdl, k) for k in names}
    globals().update(loads)



from __future__ import annotations

__doc__ = """
This module includes a basic core protocols used in many place throughout.
the library. For specific behaviors, protocols will be specified in the
module/protocol.py files.
"""
__all__ = ["_Loggable"]

from collections.abc import Callable, Generator
from typing import Any, Protocol
import logging

# Lazy-callable function takes generators as input and returns a generator
_LazyCallable = Callable[[Generator[Any]], Generator[Any]]  # FIXME


class _Loggable(Protocol):
    """
    A protocol for a logger policy.
    """

    @property
    def logger(self) -> logging.Logger: ...


class _Jsonable(Protocol):
    def to_json(self) -> dict[str, Any]: ...

    # TODO: need more features to switch the I/O of the logger or MPI-aware logging.



# from typing import TypeAlias

import neo
import numpy as np


SignalType = (
    np.ndarray | neo.core.AnalogSignal
)  # Shape should be [signal_length, n_channel]
SpikestampsType = np.ndarray | neo.core.SpikeTrain



from __future__ import annotations

__doc__ = """
Here, we define the behavior of basic operator class, and useful mixin classes that can
be used to create new operators that conform to required behaviors.
"""
__all__ = [
    "DataLoaderMixin",
    "DataNodeMixin",
    "OperatorMixin",
]

from typing import TYPE_CHECKING, Any, cast
from collections.abc import Generator
from typing_extensions import Self

import pathlib


from miv.core.operator.cachable import (
    DataclassCacher,
    FunctionalCacher,
)
from miv.core.operator.callback import BaseCallbackMixin
from miv.core.operator.chainable import BaseChainingMixin
from miv.core.operator.loggable import DefaultLoggerMixin
from miv.core.operator.policy import VanillaRunner, _RunnerProtocol

if TYPE_CHECKING:
    from miv.core.datatype import DataTypes
    from miv.core.datatype.signal import Signal
    from miv.core.datatype.spikestamps import Spikestamps
    from .protocol import _Node

else:

    class _Node: ...


class DataNodeMixin(BaseChainingMixin, DefaultLoggerMixin):
    """ """

    data: DataTypes

    def output(self) -> Self:
        return self

    def run(self) -> Self:
        return self.output()


class DataLoaderMixin(BaseChainingMixin, BaseCallbackMixin, DefaultLoggerMixin):
    """ """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.runner = VanillaRunner()
        super().__init__(*args, cacher=FunctionalCacher(self), **kwargs)
        self._load_param: dict = {}

        # Attribute from upstream
        self.tag: str

    def __call__(self) -> DataTypes:
        raise NotImplementedError("Please implement __call__ method.")

    def configure_load(self, **kwargs: Any) -> None:
        """
        (Experimental Feature)
        """
        self._load_param = kwargs

    def output(self) -> Generator[DataTypes] | Spikestamps | Generator[Signal]:
        output = self.load(**self._load_param)
        return output

    def load(
        self, *args: Any, **kwargs: Any
    ) -> Generator[DataTypes] | Spikestamps | Generator[Signal]:
        raise NotImplementedError("load() method must be implemented to be DataLoader.")


class OperatorMixin(BaseChainingMixin, BaseCallbackMixin, DefaultLoggerMixin):
    """
    Behavior includes:
        - Whenever "run()" method is executed:
            1. Check if the module is cached in the same parameters. Y: Pass to 6
            2. Callback: before run
            3. Run
            4. Callback: after run
            5. Save cache
            6. Exit
        - Cache includes:
            - All results from callback
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.runner: _RunnerProtocol = VanillaRunner()
        self.analysis_path = "analysis"

        super().__init__(*args, cacher=DataclassCacher(self), **kwargs)

        # Attribute from upstream
        self.tag: str

    def __call__(self) -> DataTypes:
        raise NotImplementedError("Please implement __call__ method.")

    def __repr__(self) -> str:
        return self.tag

    def __str__(self) -> str:
        return self.tag

    def receive(self) -> list[DataTypes]:
        """
        Receive input data from each upstream operator.
        Essentially, this method recursively call upstream operators' run() method.
        """
        return [cast(_Node, node).output() for node in self.iterate_upstream()]

    def output(self) -> DataTypes:
        """
        Output viewer. If cache exist, read result from cache value.
        Otherwise, execute (__call__) the module and return the value.
        """
        cache_called = self.cacher.check_cached()
        if cache_called:
            self.logger.info(f"Using cache: {self.cacher.cache_dir}")
            loader = self.cacher.load_cached()
            output = next(loader, None)
        else:
            self.logger.info("Cache not found.")
            self.logger.info(f"Using runner: {self.runner.__class__} type.")
            args = self.receive()  # Receive data from upstream
            if len(args) == 0:
                output = self.runner(self.__call__)
            else:
                output = self.runner(self.__call__, args)

            # Callback: After-run
            self._callback_after_run(output)

            # Plotting: Only happened when cache is not called
            self._callback_plot(output, args, show=False)

        return output

    def plot(
        self, show: bool = False, save_path: str | pathlib.Path | None = None
    ) -> None:
        """
        Standalone plotting operation.
        """
        cache_called = self.cacher.check_cached()
        if not cache_called:
            raise NotImplementedError(
                "Standalone plotting is only possible if this operator has"
                "results stored in cache. Please use Pipeline(op).run() first."
            )
        loader = self.cacher.load_cached()
        output = next(loader, None)

        # Plotting: Only happened when cache is not called
        args = self.receive()  # Receive data from upstream
        self._done_flag_plot = False  # FIXME
        self._callback_plot(output, args, show=show, save_path=save_path)
